# coding: utf-8
"""
Objects to analyze and visualize the results of GWR calculations.
"""
from __future__ import annotations

import dataclasses
import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from collections.abc import Iterable
from typing import Union, Any
from monty.functools import lazy_property
from monty.string import list_strings, marquee
from monty.termcolor import cprint
from abipy.core.structure import Structure
from abipy.core.kpoints import Kpoint, KpointList, Kpath, IrredZone, has_timrev_from_kptopt
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.iotools import ETSF_Reader
from abipy.tools import duck
from abipy.tools.typing import Figure, KptSelect
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, Marker,
    set_axlims, set_ax_xylabels, set_visible, rotate_ticklabels, set_grid_legend, hspan_ax_line, Exposer)
from abipy.tools.numtools import data_from_cplx_mode
from abipy.electrons.ebands import ElectronBands, RobotWithEbands
from abipy.electrons.gw import SelfEnergy, QPState, QPList #, SigresFile, SigresRobot
from abipy.abio.robots import Robot
from abipy.abio.enums import GWR_TASK

__all__ = [
    "GwrFile",
    "GwrRobot",
    "TchimVsSus",
]


class _MyQpkindsList(list):
    """Returned by find_qpkinds."""

@dataclasses.dataclass
class MinimaxMesh:
    """
    The minimax mesh reported in the GWR file.
    """
    ntau: int              # Number of points.
    tau_mesh: np.ndarray   # tau points.
    tau_wgs: np.ndarray    # tau weights for integration.
    iw_mesh: np.ndarray    # omega points along the imag. axis.
    iw_wgs: np.ndarray     # omega weights for integration.
    cosft_wt: np.ndarray   # weights for cosine transform (tau --> omega).
    cosft_tw: np.ndarray   # weights for cosine transform (omega --> tau).
    sinft_wt: np.ndarray   # weights for sine transform (tau --> omega).

    min_transition_energy_eV: float   # Minimum transition energy.
    max_transition_energy_eV: float   # Maximum transition energy.
    eratio: float
    ft_max_err_t2w_cos: float
    ft_max_err_w2t_cos: float
    ft_max_err_t2w_sin: float
    cosft_duality_error: float
    regterm: float                    # Regularization term.

    @classmethod
    def from_ncreader(cls, reader: ETSF_Reader) -> MinimaxMesh:
        """
        Build minimax mesh from a netcdf reader.
        """
        d = {}
        for field in dataclasses.fields(cls):
            name = field.name
            if name == "ntau":
                value = reader.read_dimvalue(name)
            else:
                value = reader.read_value(name)
                if field.type == "float":
                    value = float(value)
            d[name] = value

        return cls(**d)

    @add_fig_kwargs
    def plot_ft_weights(self, other: MinimaxMesh, self_name="self", other_name="other",
                        with_sinft=False, fontsize=6, **kwargs):
        """
        Plot the Fourier transformt weights of two minimax meshes (self and other)
        """
        if self.ntau != other.ntau:
            raise ValueError("Cannot compare minimax meshes with different ntau")

        import matplotlib.pyplot as plt
        nrows, ncols = (4, 2) if with_sinft else (3, 2)
        fig, ax_mat = plt.subplots(nrows=nrows, ncols=ncols,
                                   sharex=False, sharey=False, squeeze=False,
                                   figsize=(12, 8),
                                   #subplot_kw={'xticks': [], 'yticks': []},
        )

        I_mat = np.eye(self.ntau)
        select_irow = {
            0: [(self.cosft_wt @ self.cosft_tw) - I_mat,
                (other.cosft_wt @ other.cosft_tw) - I_mat], # , other.cosft_wt @ other.cosft_tw],
            1: [self.cosft_wt, other.cosft_wt], # self.cosft_wt - other.cosft_wt],
            2: [self.cosft_tw, other.cosft_tw], # self.cosft_tw - other.cosft_tw],
            3: [self.sinft_tw, other.sinft_tw], # self.sinft_tw - other.sinft_tw],
        }

        label_irow = {
            0: [f"(cosft_wt @ cosft_tw) - I ({self_name})", f"(cosft_wt @ cosft_tw) - I ({other_name})"],
            1: [f"cosft_wt ({self_name})", f"cosft_wt ({other_name})"],
            2: [f"cosft_tw ({self_name})", f"cosft_tw ({other_name})"],
            3: [f"sinft_tw ({self_name})", f"sinft_tw ({other_name})"],
        }

        for irow in range(nrows):
            for iax, (ax, data, label) in enumerate(zip(ax_mat[irow], select_irow[irow], label_irow[irow])):
                im = ax.matshow(data, cmap='seismic')
                fig.colorbar(im, ax=ax)
                ax.set_title(label, fontsize=fontsize)

        return fig


class GwrFile(AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This object provides an high-level interface to the GWR.nc file produced by the GWR code.
    """

    # Markers used for up/down bands.
    marker_spin = {0: "^", 1: "v"}

    color_spin = {0: "k", 1: "r"}

    @classmethod
    def from_file(cls, filepath: str) -> GwrFile:
        """Initialize an instance from filepath."""
        return cls(filepath)

    def __init__(self, filepath: str):
        """Read data from the netcdf filepath."""
        super().__init__(filepath)
        self.r = self.reader = GwrReader(self.filepath)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.r.structure

    @property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| with the KS energies."""
        return self.r.bands

    @property
    def sigma_kpoints(self) -> KpointList:
        """The k-points where QP corrections have been calculated."""
        return self.r.sigma_kpoints

    @property
    def nkcalc(self) -> int:
        """Number of k-points in sigma_kpoints"""
        return self.r.nkcalc

    @lazy_property
    def ks_dirgaps(self) -> np.ndarray:
       """KS direct gaps in eV. Shape: [nsppol, nkcalc]"""
       return self.r.read_value("ks_gaps") * abu.Ha_eV

    @lazy_property
    def qpz0_dirgaps(self) -> np.ndarray:
       """
       QP direct gaps in eV computed with the Z factor at the KS energy
       Shape: [nsppol, nkcalc]
       """
       return self.r.read_value("qpz_gaps") * abu.Ha_eV

    @lazy_property
    def minimax_mesh(self) -> MinimaxMesh:
        """Object storing minimax mesh and weights."""
        return MinimaxMesh.from_ncreader(self.r)

    @lazy_property
    def qplist_spin(self) -> tuple[QPList]:
        """Tuple of QPList objects indexed by spin."""
        return self.r.read_allqps()

    def find_qpkinds(self, qp_kpoints) -> _MyQpkindsList:
        """
        Find kpoints for QP corrections from user input.
        Return list of (kpt, ikcalc) tuples where kpt is a |Kpoint| and
        ikcalc is the index in the nkcalc array..
        """
        if isinstance(qp_kpoints, _MyQpkindsList):
            return qp_kpoints

        if isinstance(qp_kpoints, Kpoint):
            qp_kpoints = [qp_kpoints]

        if qp_kpoints is None or (duck.is_string(qp_kpoints) and qp_kpoints == "all"):
            # qp_kpoints in (None, "all")
            items = self.sigma_kpoints, list(range(self.nkcalc))

        elif duck.is_intlike(qp_kpoints):
            # qp_kpoints = 1
            ikc = int(qp_kpoints)
            items = [self.sigma_kpoints[ikc]], [ikc]

        elif isinstance(qp_kpoints, Iterable):
            # either [0, 1] or [[0, 0.5, 0]]
            # note possible ambiguity with [0, 0, 0] that is treated as list of integers.
            if duck.is_intlike(qp_kpoints[0]):
                ik_list = duck.list_ints(qp_kpoints)
                items = [self.sigma_kpoints[ikc] for ikc in ik_list], ik_list
            else:
                ik_list = [self.r.kpt2ikcalc(kpt) for kpt in qp_kpoints]
                qp_kpoints = [self.sigma_kpoints[ikc] for ikc in ik_list]
                items = qp_kpoints, ik_list
        else:
            raise TypeError("Don't know how to interpret `%s`" % (type(qp_kpoints)))

        # Check indices
        errors = []
        eapp = errors.append
        for ikc in items[1]:
            if ikc >= self.nkcalc:
                eapp("K-point index %d >= nkcalc %d, check input qp_kpoints" % (ikc, self.nkcalc))
        if errors:
            raise ValueError("\n".join(errors))

        return _MyQpkindsList(zip(items[0], items[1]))

    @lazy_property
    def params(self) -> dict:
        """
        dict with parameters that might be subject to convergence studies e.g ecuteps.
        """
        minimax_mesh = self.minimax_mesh
        r = self.r
        return dict(
            gwr_ntau=r.read_dimvalue("ntau"),
            nband=self.ebands.nband,
            ecuteps=r.read_value("ecuteps"),
            ecutsigx=r.read_value("ecutsigx"),
            ecut=r.read_value("ecut"),
            gwr_boxcutmin=r.read_value("gwr_boxcutmin"),
            nkpt=self.ebands.nkpt,
            symchi=r.read_value("symchi"),
            symsigma=r.read_value("symsigma"),
            regterm=minimax_mesh.regterm,
        )

    def close(self) -> None:
        """Close the netcdf file."""
        self.r.close()

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosity level ``verbose``."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(title="KS Electron Bands", with_structure=False))
        app("")

        # GWR section.

        app(marquee("GWR parameters", mark="="))
        app("gwr_task: %s" % self.r.gwr_task)

        if self.r.gwr_task == GWR_TASK.RPA_ENERGY:
            pass
            #d = self.get_rpa_ene_dict()

        else:
            app("Number of k-points in Sigma_{nk}: %d" % (len(self.r.sigma_kpoints)))
            app("Number of bands included in e-e self-energy sum: %d" % (self.nband))
            keys = self.params.keys() if verbose else ["ecuteps", "ecutsigx", "ecut", "gwr_boxcutmin"]
            for k in keys:
                app("%s: %s" % (k, self.params[k]))
            if verbose:
                app(marquee("k-points in Sigma_nk", mark="="))
                app(self.r.sigma_kpoints.to_string(title=None, pre_string="\t"))
            app("")

            app(marquee("QP direct gaps in eV", mark="="))
            app(str(self.get_dirgaps_dataframe(with_params=False)))
            app("")

        #app(str(self.minimax_mesh))

        return "\n".join(lines)

    #def get_qpgap(self, spin, kpoint, with_ksgap=False):
    #    """
    #    Return the QP gap in eV at the given (spin, kpoint)
    #    """
    #    ik = self.reader.kpt2fileindex(kpoint)
    #    if not with_ksgap:
    #        return self.qpgaps[spin, ik]
    #    else:
    #        return self.qpgaps[spin, ik], self.ksgaps[spin, ik]

    def get_dirgaps_dataframe(self, with_params: bool=True, with_geo: bool=False) -> pd.DataFrame:
        """
        Return a pandas DataFrame with the QP direct gaps in eV.

        Args:
            with_params: True if GWR parameters should be included.
            with_geo: True if geometry info should be included.
        """
        d = {}
        d["kpoint"] = [k.frac_coords for k in self.sigma_kpoints] * self.nsppol
        d["kname"] = [k.name for k in self.sigma_kpoints] * self.nsppol
        d["ks_dirgaps"] = self.ks_dirgaps.ravel()
        d["qpz0_dirgaps"] = self.qpz0_dirgaps.ravel()
        #d["qp_pade_dirgaps"] = self.qp_pade_dirgaps.ravel()
        d["spin"] = [0] * len(self.sigma_kpoints)
        if self.nsppol == 2: d["spin"].extend([1] * len(self.sigma_kpoints))

        if with_params:
            for k, v in self.params.items():
                 d[k] = [v] * len(self.sigma_kpoints) * self.nsppol

        if with_geo:
            d.update(**self.structure.get_dict4pandas(with_spglib=True))

        return pd.DataFrame(d)

    def get_dataframe_sk(self, spin: int, kpoint: KptSelect, index=None,
                         ignore_imag=False, with_params=True, with_geo=False) -> pd.Dataframe:
        """
        Returns |pandas-DataFrame| with the QP results for the given (spin, k-point).

        Args:
            spin:
            kpoint:
            index:
            ignore_imag: Only real part is returned if ``ignore_imag``.
            with_params: True if GWR parameters should be included.
            with_geo: True if geometry info should be included.
        """
        rows, bands = [], []

        if with_geo:
            geo_dict = self.structure.get_dict4pandas(with_spglib=True)

        qp_list = self.r.read_qplist_sk(spin, kpoint, ignore_imag=ignore_imag)
        for qp in qp_list:
            bands.append(qp.band)
            d = qp.as_dict()

            # Add other entries that may be useful when comparing different calculations.
            if with_params:
                d.update(self.params)
            if with_geo:
                d.update(**geo_dict)

            rows.append(d)

        index = len(bands) * [index] if index is not None else bands
        return pd.DataFrame(rows, index=index, columns=list(rows[0].keys()))

    def _get_include_bands(self, include_bands: Any, spin: int) -> Union[set, None]:
        """
        Helper function to return include_bands for given spin.
        """
        if include_bands is None:
            return None

        if duck.is_listlike(include_bands) or isinstance(include_bands, set):
            return set(include_bands)

        if duck.is_string(include_bands):
            if include_bands == "all":
                return None
            if include_bands == "gap":
                lumo_band = self.ebands.lumos[spin].band
                homo_band = self.ebands.homos[spin].band
                return set(range(homo_band, lumo_band + 1))

        raise ValueError(f"Invalid value for include_bands: {include_bands}")

    def get_rpa_ene_dict(self) -> dict:
        """
        Read RPA energies computed for different ecut_chi (all in Ha)
        """
        # nctkarr_t("ecut_chi", "dp", "ncut")
        # nctkarr_t("ec_rpa_ecut", "dp", "ncut")
        d = {}
        for key in ("ecut_chi", "ec_rpa_ecut", "ec_mp2_ecut"):
            d[key] = self.r.read_value(key)

        # Extrapolate value for ecut --> oo.
        xs = d["ecut_chi"] ** (-3/2.0)
        for key in ("ec_rpa_ecut", "ec_mp2_ecut"):
            ys = d[key]
            coef = np.polyfit(xs, ys, 1)
            # poly1d_fn is a function which takes in x and returns an estimate for y
            poly1d_fn = np.poly1d(coef)
            extrap_value = poly1d_fn(0.0)
            #print(f"{extrap_value=}")
            d[key + "_inf"] = extrap_value

        #print(d)
        return d

    @add_fig_kwargs
    def plot_sigma_imag_axis(self, kpoint: KptSelect,
                             spin=0, include_bands="gap", with_tau=True,
                             fontsize=8, ax_mat=None, **kwargs) -> Figure:
        """
        Plot Sigma_nk(iw) along the imaginary axis for given k-point, spin and list of bands.

        Args:
            kpoint: k-point in self-energy. Accepts |Kpoint|, vector or index.
            spin: Spin index.
            include_bands: List of bands to include. None means all.
            with_tau:
            fontsize: Legend and title fontsize.
            ax_mat:

        Returns: |matplotlib-Figure|
        """
        nrows, ncols = (2, 2) if with_tau else (1, 2)
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=not with_tau,
                                               sharey=False, squeeze=False)
        ax_mat = np.array(ax_mat)

        # Read Sigma_nk in sigma_of_band
        ikcalc, kpoint = self.r.get_ikcalc_kpoint(kpoint)
        include_bands = self._get_include_bands(include_bands, spin)
        sigma_of_band = self.r.read_sigma_bdict_sikcalc(spin, ikcalc, include_bands)

        # Plot Sigmac_nk(iw)
        re_ax, im_ax = ax_list = ax_mat[0]
        style = dict(marker="o")
        for band, sigma in sigma_of_band.items():
            sigma.c_iw.plot_ax(re_ax, cplx_mode="re", label=f"Re band: {band}", **style)
            sigma.c_iw.plot_ax(im_ax, cplx_mode="im", label=f"Im band: {band}", **style)

        re_ax.set_ylabel(r"$\Re{\Sigma_c}(i\omega)$ (eV)")
        im_ax.set_ylabel(r"$\Im{\Sigma_c}(i\omega)$ (eV)")
        set_grid_legend(ax_list, fontsize, xlabel=r"$i\omega$ (eV)")

        if with_tau:
            # Plot Sigmac_nk(itau)
            re_ax, im_ax = ax_list = ax_mat[1]
            style = dict(marker="o")
            for band, sigma in sigma_of_band.items():
                sigma.c_tau.plot_ax(re_ax, cplx_mode="re", label=f"Re band: {band}", **style)
                sigma.c_tau.plot_ax(im_ax, cplx_mode="im", label=f"Im band: {band}", **style)

            re_ax.set_ylabel(r"$\Re{\Sigma_c}(i\tau)$ (eV)")
            im_ax.set_ylabel(r"$\Im{\Sigma_c}(i\tau)$ (eV)")
            set_grid_legend(ax_list, fontsize, xlabel=r"$i\tau$ (a.u.)")

        fig.suptitle(r"$\Sigma_{nk}$" +  f" at k-point: {kpoint}, spin: {spin}", fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_sigma_real_axis(self, kpoint: KptSelect, spin=0, include_bands="gap",
                             fontsize=8, ax_mat=None, **kwargs) -> Figure:
        """
        Plot Sigma_nk(w) along the real-axis for given k-point, spin and set of bands.

        Args:
            kpoint: k-point in self-energy. Accepts |Kpoint|, vector or index.
            spin: Spin index.
            include_bands: List of bands to include. None means all.
            fontsize: Legend and title fontsize.
            ax_mat:
        """
        nrows, ncols = 1, 2
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        ax_mat = np.array(ax_mat)

        ikcalc, kpoint = self.r.get_ikcalc_kpoint(kpoint)
        include_bands = self._get_include_bands(include_bands, spin)
        sigma_of_band = self.r.read_sigma_bdict_sikcalc(spin, ikcalc, include_bands)
        nwr = self.r.nwr

        ax_list = re_ax, im_ax = ax_mat[0]
        # Show KS gap as filled area.
        self.ebands.add_fundgap_span(ax_list, spin)

        for band, sigma in sigma_of_band.items():
            ib = band - self.r.min_bstart
            e0 = self.r.e0_kcalc[spin, ikcalc, ib]

            ys = sigma.xc.values.real
            l = sigma.xc.plot_ax(re_ax, cplx_mode="re", label=f"Re band: {band}")
            point_style = dict(color=l[0].get_color(), marker="^", markersize=5.0)
            re_ax.plot(e0, ys[nwr//2], **point_style)

            ys = sigma.xc.values.imag
            l = sigma.xc.plot_ax(im_ax, cplx_mode="im", label=f"Im band: {band}")
            point_style = dict(color=l[0].get_color(), marker="^", markersize=5.0)
            im_ax.plot(e0, ys[nwr//2], **point_style)

        re_ax.set_ylabel(r"$\Re{\Sigma_{xc}(\omega)}$ (eV)")
        im_ax.set_ylabel(r"$\Im{\Sigma_{xc}(\omega)}$ (eV)")
        set_grid_legend(ax_list, fontsize=fontsize, xlabel=r"$\omega$ (eV)")

        fig.suptitle(r"$\Sigma_{nk}(\omega)$" +  f" at k-point: {kpoint}", fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_qps_vs_e0(self, with_fields="all", exclude_fields=None, e0="fermie",
                       xlims=None, sharey=False, ax_list=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot QP results stored in the GWR file as function of the KS energy.

        Args:
            with_fields: The names of the qp attributes to plot as function of eKS.
                Accepts: List of strings or string with tokens separated by blanks.
                See :class:`QPState` for the list of available fields.
            exclude_fields: Similar to ``with_fields`` but excludes fields
            e0: Option used to define the zero of energy in the band structure plot. Possible values:
                - `fermie`: shift all eigenvalues to have zero energy at the Fermi energy (`self.fermie`).
                -  Number e.g e0=0.5: shift all eigenvalues to have zero energy at 0.5 eV
                -  None: Don't shift energies, equivalent to e0=0
            ax_list: List of |matplotlib-Axes| for plot. If None, new figure is produced.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            sharey: True if y-axis should be shared.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        with_fields = QPState.get_fields_for_plot(with_fields, exclude_fields)

        # Because qplist does not have the fermi level.
        fermie = self.ebands.get_e0(e0) if e0 is not None else None
        for spin in range(self.nsppol):
            fig = self.qplist_spin[spin].plot_qps_vs_e0(
                with_fields=with_fields, exclude_fields=exclude_fields, fermie=fermie,
                xlims=xlims, sharey=sharey, ax_list=ax_list, fontsize=fontsize,
                marker=self.marker_spin[spin], show=False, **kwargs)
            ax_list = fig.axes

        return fig

    @add_fig_kwargs
    def plot_spectral_functions(self, include_bands=None, ax_list=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the spectral function A_{nk}(w) for all k-points, bands and
        spins available in the GWR file.

        Args:
            include_bands: List of bands to include. None means all.
            ax_list:
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        # Build grid of plots.
        nrows, ncols = len(self.sigma_kpoints), 1
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = np.array(ax_list).ravel()

        for ikcalc, (kcalc, ax) in enumerate(zip(self.sigma_kpoints, ax_list)):
            for spin in range(self.nsppol):
                include_bands_ks = self._get_include_bands(include_bands, spin)
                sigma_of_band = self.r.read_sigma_bdict_sikcalc(spin, ikcalc, include_bands_ks)

                for band, sigma in sigma_of_band.items():
                    label = r"$A(\omega)$: band: %d, spin: %d" % (band, spin)
                    l = sigma.plot_ax(ax, what="a", label=label, fontsize=fontsize)
                    # Show position of the KS energy as vertical line.
                    ib = band - self.r.min_bstart
                    ax.axvline(self.r.e0_kcalc[spin, ikcalc, ib],
                               lw=1, color=l[0].get_color(), ls="--")

            # Show KS gap as filled area.
            self.ebands.add_fundgap_span(ax, spin)

            ax.set_xlabel(r"$\omega$ (eV)")
            ax.set_ylabel(r"$A(\omega)$ (1/eV)")
            ax.set_title("k-point: %s" % repr(kcalc), fontsize=fontsize)

        return fig

    #@add_fig_kwargs
    #def plot_sparsity(self, what, origin="lower", **kwargs):
    #   x_mat
    #   ax.spy(mat, precision=0.1, markersize=5, origin=origin)

    #@add_fig_kwargs
    #def plot_mat(self, what, origin="lower", **kwargs):
    #   x_mat
    #   ax.spy(mat, precision=0.1, markersize=5, origin=origin)

    #def get_panel(self, **kwargs):
    #    """
    #    Build panel with widgets to interact with the GWR.nc either in a notebook or in panel app.
    #    """
    #    from abipy.panels.gwr import GwrFilePanel
    #    return GwrFilePanel(self).get_panel(**kwargs)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        verbose = kwargs.pop("verbose", 0)

        include_bands = "all" if verbose else "gaps"
        yield self.plot_spectral_functions(include_bands=include_bands, show=False)

        # TODO
        for spin in range(self.nsppol):
            for ik, kpoint in enumerate(self.sigma_kpoints):
                kws = dict(spin=spin, include_bands = "gaps", show=False)
                yield self.plot_sigma_imag_axis(ik, **kws)
                yield self.plot_sigma_real_axis(ik, **kws)

    def write_notebook(self, nbpath=None, title=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=title)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class GwrReader(ETSF_Reader):
    r"""
    This object provides methods to read data from the GWR.nc file.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GwrReader
    """

    def __init__(self, path: str):
        super().__init__(path)

        # Save important quantities needed to simplify the API.
        self.bands = ElectronBands.from_file(path)
        self.structure = self.read_structure()

        self.nsppol = self.bands.nsppol
        self.nwr = self.read_dimvalue("nwr")
        self.nkcalc = self.read_dimvalue("nkcalc")
        self.smat_bsize1 = self.read_dimvalue("smat_bsize1")
        self.smat_bsize2 = self.read_dimvalue("smat_bsize2")
        char_list = self.read_value("gwr_task")
        for i, ch in enumerate(reversed(char_list)):
            if not ch: break
        i = len(char_list) + i
        self.gwr_task = "".join(c.decode("utf-8") for c in char_list[:i])
        self.sig_diago = bool(self.read_value("sig_diago"))

        # The k-points where QP corrections have been calculated.
        kcalc_red_coords = self.read_value("kcalc")
        self.sigma_kpoints = KpointList(self.structure.reciprocal_lattice, kcalc_red_coords)
        # Find k-point name
        for kpoint in self.sigma_kpoints:
            kpoint.set_name(self.structure.findname_in_hsym_stars(kpoint))

        # Read mapping kcalc --> IBZ and convert to C indexing.
        # nctkarr_t("kcalc2ibz", "int", "nkcalc, six") &
        self.kcalc2ibz = self.read_variable("kcalc2ibz")[0,:] - 1

        # Note conversion between Fortran and python convention.
        self.bstart_sk = self.read_value("bstart_ks") - 1
        self.bstop_sk = self.read_value("bstop_ks")
        # min band index for GW corrections over spins and k-points
        self.min_bstart = np.min(self.bstart_sk)

    @lazy_property
    def iw_mesh(self) -> np.ndarray:
        """Frequency mesh in eV along the imaginary axis."""
        return self.read_value("iw_mesh") * abu.Ha_eV

    @lazy_property
    def tau_mesh(self) -> np.ndarray:
        """Tau mesh in a.u."""
        return self.read_value("tau_mesh")

    @lazy_property
    def wr_step(self) -> float:
        """Frequency-mesh along the real axis in eV."""
        return self.read_value("wr_step") * abu.Ha_eV

    @lazy_property
    def e0_kcalc(self) -> np.ndarray:
        # nctkarr_t("e0_kcalc", "dp", "smat_bsize1, nkcalc, nsppol"), &
        return self.read_value("e0_kcalc") * abu.Ha_eV

    def get_ikcalc_kpoint(self, kpoint) -> tuple(int, Kpoint):
        """
        Return the ikcalc index and the Kpoint
        """
        ikcalc = self.kpt2ikcalc(kpoint)
        kpoint = self.sigma_kpoints[ikcalc]
        return ikcalc, kpoint

    def kpt2ikcalc(self, kpoint) -> int:
        """
        Returns the index of the k-point in the sigma_kpoints array.
        Used to access data in the arrays that are dimensioned [0:nkcalc]
        """
        if duck.is_intlike(kpoint):
            return int(kpoint)
        else:
            return self.sigma_kpoints.index(kpoint)

    def get_wr_mesh(self, e0: float) -> np.ndarray:
        """
        The frequency mesh in eV is linear and centered around KS e0.
        """
        nwr = self.nwr
        return np.linspace(start=e0 - self.wr_step * (nwr // 2),
                           stop=e0 + self.wr_step * (nwr // 2),  num=nwr)

    def read_sigee_skb(self, spin, kpoint, band) -> SelfEnergy:
        """"
        Read self-energy for (spin, kpoint, band).
        """
        ikcalc, kpoint = self.get_ikcalc_kpoint(kpoint)
        ib = band - self.min_bstart
        ib2 = 0 if self.sig_diago else ib

        e0 = self.e0_kcalc[spin, ikcalc, ib]
        wmesh = self.get_wr_mesh(e0)

        # nctkarr_t("sigxc_rw_diag", "dp", "two, nwr, smat_bsize1, nkcalc, nsppol"), &
        sigxc_values = self.read_variable("sigxc_rw_diag")[spin,ikcalc,ib,:,:] * abu.Ha_eV
        sigxc_values = sigxc_values[:,0] + 1j *sigxc_values[:,1]

        # nctkarr_t("spfunc_diag", "dp", "nwr, smat_bsize1, nkcalc, nsppol") &
        spf_values = self.read_variable("spfunc_diag")[spin,ikcalc,ib,:] / abu.Ha_eV

        # nctkarr_t("sigc_iw_mat", "dp", "two, ntau, smat_bsize1, smat_bsize2, nkcalc, nsppol"), &
        sigc_iw = self.read_value("sigc_iw_mat", cmode="c") * abu.Ha_eV
        c_iw_values = sigc_iw[spin,ikcalc,ib2,ib]

        # nctkarr_t("sigc_it_mat", "dp", "two, two, ntau, smat_bsize1, smat_bsize2, nkcalc, nsppol"), &
        sigc_tau = self.read_value("sigc_it_mat", cmode="c") * abu.Ha_eV
        c_tau_pm = sigc_tau[spin,ikcalc,ib2,ib]
        tau_mp_mesh = np.concatenate((-self.tau_mesh[::-1], self.tau_mesh))
        c_tau_mp_values = np.concatenate((c_tau_pm[::-1,1], c_tau_pm[:,0]))

        return SelfEnergy(spin, kpoint, band, wmesh, sigxc_values, spf_values,
                          iw_mesh=self.iw_mesh, c_iw_values=c_iw_values,
                          tau_mp_mesh=tau_mp_mesh, c_tau_mp_values=c_tau_mp_values)

    def read_sigma_bdict_sikcalc(self, spin: int, ikcalc: int, include_bands) -> dict[int, SelfEnergy]:
        """
        Return dict of self-energy objects for given (spin, ikcalc) indexed by the band index.
        """
        sigma_of_band = {}
        for band in range(self.bstart_sk[spin, ikcalc], self.bstop_sk[spin, ikcalc]):
            if include_bands and band not in include_bands: continue
            sigma_of_band[band] = self.read_sigee_skb(spin, ikcalc, band)
        return sigma_of_band

    def read_allqps(self, ignore_imag=False) -> tuple[QPList]:
        """
        Return list with ``nsppol`` items. Each item is a :class:`QPList` with the QP results

        Args:
            ignore_imag: Only real part is returned if ``ignore_imag``.
        """
        qps_spin = self.nsppol * [None]

        for spin in range(self.nsppol):
            qps = []
            for kpoint in self.sigma_kpoints:
                ikcalc = self.kpt2ikcalc(kpoint)
                qps.extend(self.read_qplist_sk(spin, kpoint, ignore_imag=ignore_imag))

            qps_spin[spin] = QPList(qps)

        return tuple(qps_spin)

    def read_qplist_sk(self, spin, kpoint, ignore_imag=False) -> QPList:
        """
        Read and return a QPList object for the given spin, kpoint.

        Args:
            ignore_imag: Only real part is returned if ``ignore_imag``.
        """
        ikcalc, kpoint = self.get_ikcalc_kpoint(kpoint)

        def ri(a):
            return np.real(a) if ignore_imag else a

        # TODO: Finalize the implementation.
        #sigxme = sigx_mat
        sigxme = 0.0
        #self._sigxme[spin, ikcalc, ib],
        qp_list = QPList()
        for band in range(self.bstart_sk[spin, ikcalc], self.bstop_sk[spin, ikcalc]):
            ib = band - self.min_bstart

            qpe = self.read_variable("qpz_ene")[spin, ikcalc, ib] * abu.Ha_meV
            qpe = qpe[0] + 1j*qpe[1]

            ze0 = self.read_variable("ze0_kcalc")[spin, ikcalc, ib]
            ze0 = ze0[0] + 1j*ze0[1]

            # TODO Finalize the implementation
            qp_list.append(QPState(
                spin=spin,
                kpoint=kpoint,
                band=band,
                e0=self.e0_kcalc[spin, ikcalc, ib],
                qpe=ri(qpe),
                qpe_diago=0.0,
                #vxcme=self._vxcme[spin, ikcalc, ib],
                vxcme=0.0,
                sigxme=sigxme,
                #sigcmee0=ri(self._sigcmee0[spin, ikcalc, ib]),
                sigcmee0=0.0,
                vUme=0.0,
                ze0=ri(ze0),
            ))
        return qp_list


class GwrRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple GWR.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GwrRobot
    """
    # Try to have API similar to SigresRobot
    EXT = "GWR"

    # matplotlib option to fill convergence window.
    HATCH = "/"

    def __init__(self, *args):
        super().__init__(*args)
        if len(self.abifiles) in (0, 1): return

        # Check dimensions and self-energy states and issue warning.
        warns = []; wapp = warns.append
        nc0 : GwrFile = self.abifiles[0]
        same_nsppol, same_nkcalc = True, True
        if any(nc.nsppol != nc0.nsppol for nc in self.abifiles):
            same_nsppol = False
            wapp("Comparing ncfiles with different values of nsppol.")
        if any(nc.r.nkcalc != nc0.r.nkcalc for nc in self.abifiles):
            same_nkcalc = False
            wapp("Comparing ncfiles with different number of k-points in self-energy. Doh!")

        if same_nsppol and same_nkcalc:
            # FIXME
            # Different values of bstart_ks are difficult to handle
            # Because the high-level API assumes an absolute global index
            # Should decide how to treat this case: either raise or interpret band as an absolute band index.
            if any(np.any(nc.r.bstart_sk != nc0.r.bstart_sk) for nc in self.abifiles):
                wapp("Comparing ncfiles with different values of bstart_sk")
            if any(np.any(nc.r.bstop_sk != nc0.r.bstop_sk) for nc in self.abifiles):
                wapp("Comparing ncfiles with different values of bstop_sk")

        if warns:
            for w in warns:
                cprint(w, color="yellow")

    def _check_dims_and_params(self) -> None:
        """
        Test nsppol, sigma_kpoints.
        """
        if not len(self.abifiles) > 1:
            return

        nc0: GwrFile = self.abifiles[0]
        errors = []
        eapp = errors.append

        if any(nc.nsppol != nc0.nsppol for nc in self.abifiles[1:]):
            eapp("Files with different values of `nsppol`")

        if any(nc.nkcalc != nc0.nkcalc for nc in self.abifiles[1:]):
            cprint("Files with different values of `nkcalc`", color="yellow")

        for nc in self.abifiles[1:]:
            for k0, k1 in zip(nc0.sigma_kpoints, nc.sigma_kpoints):
                if k0 != k1:
                    cprint("Files with different values of `sigma_kpoints`\n"+
                           "Specify the kpoint via reduced coordinates and not via the index", "yellow")
                    break

        if errors:
            raise ValueError("Cannot compare multiple GWR.nc files. Reason:\n %s" % "\n".join(errors))

    def get_dataframe_sk(self, spin, kpoint, with_params=True, ignore_imag=False) -> pd.DataFrame:
        """
        Return |pandas-Dataframe| with QP results for this spin, k-point

        Args:
            spin: Spin index
            kpoint: K-point in self-energy. Accepts |Kpoint|, vector or index.
            with_params: True to add convergence parameters.
            ignore_imag: only real part is returned if ``ignore_imag``.
        """
        df_list = []; app = df_list.append
        for label, ncfile in self.items():
            df = ncfile.get_dataframe_sk(spin, kpoint, index=None,
                                         with_params=with_params, ignore_imag=ignore_imag)
            app(df)

        return pd.concat(df_list)

    def get_dirgaps_dataframe(self, sortby="kname", with_params=True) -> pd.DataFrame:
        """
        Returns |pandas-DataFrame| with QP direct gaps for all the files treated by the GWR robot.

        Args:
            sortby: Name to sort by.
            with_params: False to exclude calculation parameters from the dataframe.
        """
        with_geo = self.has_different_structures()

        df_list = []; app = df_list.append
        for _, ncfile in self.items():
            app(ncfile.get_dirgaps_dataframe(with_params=with_params, with_geo=with_geo))

        df = pd.concat(df_list)
        if sortby and sortby in df: df = df.sort_values(sortby)
        return df

    def get_dataframe(self, sortby="kname", with_params=True, ignore_imag=False) -> pd.DataFrame:
        """
        Return |pandas-Dataframe| with QP results for all k-points, bands and spins
        present in the files treated by the GWR robot.

        Args:
            sortby: Name to sort by.
            with_params: True to add parameters.
            ignore_imag: only real part is returned if ``ignore_imag``.
        """
        df_list = []; app = df_list.append
        for _, ncfile in self.items():
            for spin in range(ncfile.nsppol):
                for ikc, _ in enumerate(ncfile.sigma_kpoints):
                    app(ncfile.get_dataframe_sk(spin, ikc, with_params=with_params,
                                                ignore_imag=ignore_imag))

        df = pd.concat(df_list)
        if sortby and sortby in df: df = df.sort_values(sortby)
        return df

    def get_rpa_ene_dataframe(self, with_params=True) -> pd.DataFrame:
        """
        Return |pandas-Dataframe| with RPA energies for all the files
        treated by the GWR robot.

        Args:
            with_params: True to add parameters.
        """
        keys = ["ec_rpa_ecut_inf", "ec_mp2_ecut_inf"]
        dict_list = []
        for _, ncfile in self.items():
            d = ncfile.get_rpa_ene_dict()
            d = {k: d[k] for k in keys}
            if with_params:
                d.update(ncfile.params)
            dict_list.append(d)

        df = pd.DataFrame(dict_list)
        #if sortby and sortby in df: df = df.sort_values(sortby)
        return df

    @add_fig_kwargs
    def plot_selfenergy_conv(self, spin, kpoint, band, axis="wreal", sortby=None, hue=None,
                             colormap="viridis", xlims=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the e-e self-energy wrt to the ``sortby`` parameter.
        Values can be optionally grouped by `hue`.

        Args:
            spin: Spin index.
            kpoint: K-point in self-energy. Accepts |Kpoint|, vector or index.
            band: Band index.
            axis: "wreal": to plot Sigma(w) and A(w) along the real axis.
                  "wimag": to plot Sigma(iw)
                  "tau": to plot Sigma(itau)) along the imag axis.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            colormap: matplotlib color map.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap(colormap)

        # Make sure nsppol and sigma_kpoints are consistent.
        self._check_dims_and_params()
        ebands0 = self.abifiles[0].ebands

        if hue is None:
            # Build grid depends on axis.
            nrows = {"wreal": 3, "wimag": 2, "tau": 2}[axis]
            ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=1,
                                                    sharex=True, sharey=False, squeeze=False)
            ax_list = np.array(ax_list).ravel()

            lnp_list = self.sortby(sortby)
            for ix, (nclabel, ncfile, param) in enumerate(lnp_list):
                label = "%s: %s" % (self._get_label(sortby), param)
                kws = dict(label=label or nclabel, color=cmap(ix / len(lnp_list)))
                sigma = ncfile.r.read_sigee_skb(spin, kpoint, band)

                if axis == "wreal":
                    # Plot Sigma(w) along the real axis.
                    sigma.plot_reima_rw(ax_list, **kws)
                elif axis == "wimag":
                    # Plot Sigma(iw) along the imaginary axis.
                    sigma.plot_reimc_iw(ax_list, **kws)
                elif axis == "tau":
                    # Plot Sigma(itau) along the imaginary axis.
                    sigma.plot_reimc_tau(ax_list, **kws)

            if axis == "wreal": ebands0.add_fundgap_span(ax_list, spin)
            set_grid_legend(ax_list, fontsize)

        else:
            # group_and_sortby and build (3, ngroups) subplots
            groups = self.group_and_sortby(hue, sortby)
            nrows = {"wreal": 3, "wimag": 2, "tau": 2}[axis]
            ncols = len(groups)
            ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                   sharex=True, sharey=False, squeeze=False)
            for ig, g in enumerate(groups):
                subtitle = "%s: %s" % (self._get_label(hue), g.hvalue)
                ax_list = ax_mat[:,ig]
                ax_list[0].set_title(subtitle, fontsize=fontsize)

                for ix, (nclabel, ncfile, param) in enumerate(g):
                    kws = dict(label="%s: %s" % (self._get_label(sortby), param), color=cmap(ix / len(g)))
                    sigma = ncfile.r.read_sigee_skb(spin, kpoint, band)

                    if axis == "wreal":
                        lines = sigma.plot_reima_rw(ax_list, **kws)
                        if ix == 0:
                            # Show position of KS energy as vertical line.
                            ikcalc, _ = ncfile.r.get_ikcalc_kpoint(kpoint)
                            ib = band - ncfile.r.min_bstart
                            for ax, l in zip(ax_list, lines):
                                ax.axvline(ncfile.r.e0_kcalc[spin, ikcalc, ib], lw=1, color=l[0].get_color(), ls="--")

                    elif axis == "wimag":
                        # Plot Sigma_c(iw) along the imaginary axis.
                        sigma.plot_reimc_iw(ax_list, **kws)

                    elif axis == "tau":
                        # Plot Sigma(itau) along the imaginary axis.
                        sigma.plot_reimc_tau(ax_list, **kws)

                if axis == "wreal": ebands0.add_fundgap_span(ax_list, spin)
                set_grid_legend(ax_list, fontsize)
                for ax in ax_list:
                    set_axlims(ax, xlims, "x")

            if ig != 0:
                set_visible(ax_mat[:, ig], False, "ylabel")

        _, kpoint = self.abifiles[0].r.get_ikcalc_kpoint(kpoint)
        fig.suptitle(r"$\Sigma_{nk}$" + f" at k-point: {kpoint}, band: {band}, spin: {spin}",
                     fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_qpgaps_convergence(self, qp_kpoints="all", qp_type="qpz0", sortby=None, hue=None, abs_conv=0.01,
                                plot_qpmks=True, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the direct QP gaps for all the k-points and spins treated by the GWR robot.

        Args:
            qp_kpoints: List of k-points in self-energy. Accept integers (list or scalars), list of vectors,
                or "all" to plot all k-points.
            qp_type: "qpz0" for linear qp equation with Z factor computed at KS e0,
                     "otms" for on-the-mass-shell values.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            abs_conv: If not None, show absolute convergence window.
            plot_qpmks: If False, plot QP_gap, KS_gap else (QP_gap - KS_gap)
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        # Make sure nsppol and sigma_kpoints are the same.
        self._check_dims_and_params()

        nc0 = self.abifiles[0]
        nsppol = nc0.nsppol
        qpkinds = nc0.find_qpkinds(qp_kpoints)
        if len(qpkinds) > 10:
            cprint("More that 10 k-points in file. Only 10 k-points will be shown. Specify kpt index expliclty", "yellow")
            qpkinds = qpkinds[:10]

        # Build grid with (nkpt, 1) plots.
        nrows, ncols = len(qpkinds), 1
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        if hue is None:
            labels, ncfiles, xs = self.sortby(sortby, unpack=True)
        else:
            groups = self.group_and_sortby(hue, sortby)

        #if qp_type not in {"qpz0", "otms"}:
        if qp_type not in {"qpz0", }:
            raise ValueError("Invalid qp_type: %s" % qp_type)

        name = "QP dirgap" if not plot_qpmks else "QP - KS dirgap"
        name = "%s (%s)" % (name, qp_type.upper())

        for ix, ((kpt, ikc), ax_row) in enumerate(zip(qpkinds, ax_mat)):
            ax = ax_row[0]
            for spin in range(nsppol):
                ax.set_title("%s k:%s" % (name, repr(kpt)), fontsize=fontsize)

                # Extract QP dirgap for [spin, ikcalc]
                if hue is None:
                    if qp_type == "qpz0": yvals = [ncfile.qpz0_dirgaps[spin, ikc] for ncfile in ncfiles]
                    #if qp_type == "otms": yvals = [ncfile.qp_dirgaps_otms_t[spin, ikc, itemp] for ncfile in ncfiles]
                    if plot_qpmks:
                        yvals = np.array(yvals) - np.array([ncfile.ks_dirgaps[spin, ikc] for ncfile in ncfiles])

                    lines = self.plot_xvals_or_xstr_ax(ax, xs, yvals, fontsize, marker=nc0.marker_spin[spin],
                                                       **kwargs)
                    hspan_ax_line(ax, lines[0], abs_conv, self.HATCH)

                else:
                    for g in groups:
                        if qp_type == "qpz0": yvals = [ncfile.qpz0_dirgaps[spin, ikc] for ncfile in g.abifiles]
                        #if qp_type == "otms": yvals = [ncfile.qp_dirgaps_otms_t[spin, ikc, itemp] for ncfile in g.abifiles]
                        if plot_qpmks:
                            yvals = np.array(yvals) - np.array([ncfile.ks_dirgaps[spin, ikc] for ncfile in g.abifiles])

                        label = "%s: %s" % (self._get_label(hue), g.hvalue)
                        lines = ax.plot(g.xvalues, yvals, marker=nc0.marker_spin[spin], label=label)

                        hspan_ax_line(ax, lines[0], abs_conv, self.HATCH)

            ax.grid(True)
            if ix == len(qpkinds) - 1:
                ax.set_ylabel("%s (eV)" % name)
                ax.set_xlabel("%s" % self._get_label(sortby))
                if sortby is None: rotate_ticklabels(ax, 15)
            ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_qpdata_conv_skb(self, spin, kpoint, band,
                             sortby=None, hue=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the QP results for given (spin, kpoint, band).

        Args:
            spin: Spin index.
            kpoint: K-point in self-energy. Accepts |Kpoint|, vector or index.
            band: Band index.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        # Make sure that nsppol and sigma_kpoints are consistent.
        self._check_dims_and_params()

        # TODO: Add more quantities DW, Fan(0)
        # TODO: Decide how to treat complex quantities, avoid annoying ComplexWarning
        # TODO: Format for g.hvalue
        # Treat fundamental gaps
        # Quantities to plot.
        what_list = ["re_qpe", "imag_qpe", "ze0"]

        # Build grid plot.
        nrows, ncols = len(what_list), 1
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = np.array(ax_list).ravel()

        nc0: GwrFile = self.abifiles[0]
        ikc = nc0.kpt2ikcalc(kpoint)
        kpoint = nc0.sigma_kpoints[ikc]

        # Sort and read QP data.
        if hue is None:
            labels, ncfiles, params = self.sortby(sortby, unpack=True)
            qplist = [ncfile.reader.read_qp(spin, kpoint, band) for ncfile in ncfiles]
        else:
            groups = self.group_and_sortby(hue, sortby)
            qplist_group = []
            for g in groups:
                lst = [ncfile.reader.read_qp(spin, kpoint, band) for ncfile in g.abifiles]
                qplist_group.append(lst)

        for ix, (ax, what) in enumerate(zip(ax_list, what_list)):
            if hue is None:
                # Extract QP data.
                #yvals = [getattr(qp, what)[itemp] for qp in qplist]
                if not duck.is_string(params[0]):
                    ax.plot(params, yvals, marker=nc0.marker_spin[spin])
                else:
                    # Must handle list of strings in a different way.
                    xn = range(len(params))
                    ax.plot(xn, yvals, marker=nc0.marker_spin[spin])
                    ax.set_xticks(xn)
                    ax.set_xticklabels(params, fontsize=fontsize)
            else:
                for g, qplist in zip(groups, qplist_group):
                    # Extract QP data.
                    yvals = [getattr(qp, what)[itemp] for qp in qplist]
                    label = "%s: %s" % (self._get_label(hue), g.hvalue)
                    ax.plot(g.xvalues, yvals, marker=nc0.marker_spin[spin],
                            label=label if ix == 0 else None)

            ax.grid(True)
            ax.set_ylabel(what)
            if ix == len(what_list) - 1:
                ax.set_xlabel("%s" % self._get_label(sortby))
                if sortby is None: rotate_ticklabels(ax, 15)
            if ix == 0 and hue is not None:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

        if "title" not in kwargs:
            title = "QP results spin: %s, k:%s, band: %s, T = %.1f K" % (
                    spin, repr(kpoint), band, nc0.tmesh[itemp])
            fig.suptitle(title, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_qpfield_vs_e0(self, field, reim="real", function=lambda x: x, sortby=None, hue=None,
                           fontsize=8, colormap="jet", e0="fermie", **kwargs) -> Figure:
        """
        For each file in the GWR robot, plot one of the attributes of :class:`QpTempStat
        as a function of the KS energy.

        Args:
            field (str): String defining the attribute to plot.
            reim: Plot the real or imaginary part
            function: Apply a function to the results before plotting
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it is assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            colormap: matplotlib color map.
            fontsize: legend and label fontsize.
            e0: Option used to define the zero of energy in the band structure plot.

        .. note::

            For the meaning of the other arguments, see other robot methods.

        Returns: |matplotlib-Figure|
        """
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap(colormap)

        if hue is None:
            lnp_list = self.sortby(sortby)
            ax_list = None
            for i, (label, ncfile, param) in enumerate(lnp_list):
                if sortby is not None:
                    label = "%s: %s" % (self._get_label(sortby), param)
                fig = ncfile.plot_qps_vs_e0(with_fields=list_strings(field),
                    reim=reim, function=function, e0=e0, ax_list=ax_list,
                    color=cmap(i / len(lnp_list)), fontsize=fontsize,
                    label=label, show=False)
                ax_list = fig.axes
        else:
            # group_and_sortby and build (ngroups,) subplots
            groups = self.group_and_sortby(hue, sortby)
            nrows, ncols = 1, len(groups)
            ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                   sharex=True, sharey=True, squeeze=False)
            for ig, g in enumerate(groups):
                subtitle = "%s: %s" % (self._get_label(hue), g.hvalue)
                ax_mat[0, ig].set_title(subtitle, fontsize=fontsize)
                for i, (nclabel, ncfile, param) in enumerate(g):
                    fig = ncfile.plot_qps_vs_e0(with_fields=list_strings(field),
                        reim=reim, function=function,
                        e0=e0, ax_list=ax_mat[:, ig], color=cmap(i / len(g)), fontsize=fontsize,
                        label="%s: %s" % (self._get_label(sortby), param), show=False)

                if ig != 0:
                    set_visible(ax_mat[:, ig], False, "ylabel")

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        verbose = kwargs.pop("verbose", 0)
        yield self.plot_qpgaps_convergence(qp_kpoints="all", show=False)

        # Visualize the convergence of the self-energy for
        # all the k-points and the most important bands.
        nc0: GwrFile = self.abifiles[0]

        for spin in range(nc0.nsppol):
            for ikcalc in range(nc0.nkcalc):
                ik_ibz = nc0.r.kcalc2ibz[ikcalc]
                band_v = nc0.ebands.homo_sk(spin, ik_ibz).band
                band_c = nc0.ebands.lumo_sk(spin, ik_ibz).band
                for band in range(band_v, band_c + 1):
                    for axis in ("wreal", "wimag", "tau"):
                        yield self.plot_selfenergy_conv(spin, ikcalc, band, axis=axis, show=False)

    def write_notebook(self, nbpath=None, title=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=title)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            nbv.new_code_cell("robot = abilab.SigEPhRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("robot.get_params_dataframe()"),
            nbv.new_code_cell("# data = robot.get_dataframe()\ndata"),
            nbv.new_code_cell("robot.plot_qpgaps_convergence(itemp=0, sortby=None, hue=None);"),
            nbv.new_code_cell("""\
nc0 = robot.abifiles[0]
for spin in range(nc0.nsppol):
    for ikc, sigma_kpoint in enumerate(nc0.sigma_kpoints):
        for band in range(nc0.r.bstart_sk[spin, ikc], nc0.bstop_sk[spin, ikc]):
            robot.plot_qpdata_conv_skb(spin, sigma_kpoint, band, itemp=0, sortby=None, hue=None);"""),

            nbv.new_code_cell("""\
#nc0 = robot.abifiles[0]
#for spin in range(nc0.nsppol):
#    for ikc, sigma_kpoint in enumerate(nc0.sigma_kpoints):
#        for band in range(nc0.r.bstart_sk[spin, ikc], nc0.bstop_sk[spin, ikc]):
#           robot.plot_selfenergy_conv(spin, sigma_kpoint, band, itemp=0, sortby=None);"),"""),
        ])

        # Mixins.
        nb.cells.extend(self.get_baserobot_code_cells())
        nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)



#class GwrSigresComparator:
#    """
#    This object provides a high-level API to compare the results stored in
#    a GWR.nc and a SIGRES.nc file (usually produced with the AC method).
#    """
#
#    def __init__(self, gwr_filepaths, sigres_filepaths):
#        """
#        Args:
#            gwr_filepaths: Path(s) of the GWR.nc file.
#            sigres_filepaths: Path(s) of the SIGRES.nc file.
#        """
#        self.gwr_robot = GwrRobot.from_files(gwr_filepaths)
#        self.sigres_robot = SigresRobot.from_files(sigres_filepaths)
#
#    def __enter__(self):
#        return self
#
#    def __exit__(self, exc_type, exc_val, exc_tb):
#        """
#        Activated at the end of the with statement. It automatically closes the files.
#        """
#        self.gwr_robot.close()
#        self.sigres_robot.close()
#
#    def __str__(self) -> str:
#        return self.to_string()
#
#    def to_string(self, verbose: int = 0) -> str:
#        """String representation with verbosity level ``verbose``."""
#        lines = []; app = lines.append
#        app(self.gwr_robot.to_string(verbose=verbose))
#        app("")
#        app(self.sigres_robot.to_string(verbose=verbose))
#        return "\n".join(lines)


def _find_g(gg, gvecs):
    for ig, g_sus in enumerate(gvecs):
        if all(gg == g_sus): return ig
    raise ValueError(f"Cannot find g-vector: {gg}")



class TchimVsSus:
    """
    Object used to compare the polarizability computed in the GWR code with the one
    produced by the legacy algorithm based on the Adler-Wiser expression.

    Example:

        gpairs = [
            ((0, 0, 0), (1, 0, 0)),
            ((1, 0, 0), (0, 1, 0)),
        ]
        qpoint_list = [
            [0, 0, 0],
            [0.5, 0.5, 0],
        ]

        with TchimVsSus("runo_DS3_TCHIM.nc", "AW_CD/runo_DS3_SUS.nc") as o
            o.expose_qpoints_gpairs(qpoint_list, gpairs, exposer="mpl")
    """
    def __init__(self, tchim_filepath: str, sus_filepath: str):
        """
        Args:
            tchim_filepath: TCHIM filename.
            sus_filepath: SUS filename.
        """
        from abipy.electrons.scr import SusFile
        self.sus_file = SusFile(sus_filepath)
        self.tchi_reader = ETSF_Reader(tchim_filepath)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Activated at the end of the with statement. It automatically closes the files.
        """
        self.sus_file.close()
        self.tchi_reader.close()

    def find_tchim_qpoint(self, qpoint) -> tuple[int, np.ndarray]:
        """
        Find the index of the q-point in the IBZ. Return index and q-points.
        """
        tchi_qpoints = self.tchi_reader.read_variable("qibz")

        if duck.is_intlike(qpoint):
            return qpoint, tchi_qpoints

        for tchi_iq, tchi_qq in enumerate(tchi_qpoints):
            if np.all(abs(tchi_qq - qpoint) < 1e-6):
                return tchi_iq, tchi_qpoints

        raise ValueError(f"Cannot find q-point: {qpoint} in TCHIM file")

    @add_fig_kwargs
    def plot_qpoint_gpairs(self, qpoint, gpairs,
                           fontsize=8, spins=(0, 0), with_title=True, **kwargs) -> Figure:
        """
        Plot the Fourier components of the polarizability for given q-point and list of (g, g') pairs.

        Args:
            qpoint: q-point or q-point index.
            gpairs: List of (g,g') pairs
        """
        gpairs = np.array(gpairs)

        sus_reader = self.sus_file.reader
        _, sus_iq = sus_reader.find_kpoint_fileindex(qpoint)

        # int reduced_coordinates_plane_waves_dielectric_function(number_of_qpoints_dielectric_function,
        # number_of_coefficients_dielectric_function, number_of_reduced_dimensions) ;
        # SUSC file uses Gamma-centered G-spere so read at iq = 0
        susc_iwmesh = sus_reader.read_value("frequencies_dielectric_function", cmode="c").imag
        susc_gvecs = sus_reader.read_variable("reduced_coordinates_plane_waves_dielectric_function")[0]

        tchi_iq, _ = self.find_tchim_qpoint(qpoint)
        tchi_iwmesh = self.tchi_reader.read_value("iw_mesh")
        tchi_gvecs = self.tchi_reader.read_variable("gvecs")[tchi_iq]

        # Find indices of gpairs in the two files
        sus_inds = [(_find_g(g1, susc_gvecs), _find_g(g2, susc_gvecs)) for g1, g2 in gpairs]
        chi_inds = [(_find_g(g1, tchi_gvecs), _find_g(g2, tchi_gvecs)) for g1, g2 in gpairs]

        if sus_reader.path.endswith("SUS.nc"):
            sus_var_name = "polarizability"
        elif sus_reader.path.endswith("SCR.nc"):
            sus_var_name = "inverse_dielectric_function"
        else:
            raise ValueError(f"Cannot detect varname for file: {sus_reader.path}")

        # Build grid of plots.
        num_plots, ncols, nrows = 2 * len(gpairs), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        for i, ((g1, g2), (sus_ig1, sus_ig2), (chi_ig1, chi_ig2)) in enumerate(zip(gpairs, sus_inds, chi_inds)):
            ax_re, ax_im = ax_mat[i]

            # number_of_qpoints_dielectric_function, number_of_frequencies_dielectric_function,
            # number_of_spins, number_of_spins, number_of_coefficients_dielectric_function,
            # number_of_coefficients_dielectric_function, complex)
            sus_data = sus_reader.read_variable(sus_var_name)[sus_iq, :, 0, 0, sus_ig2, sus_ig1]
            sus_data = sus_data[:, 0] + 1j * sus_data[:, 1]

            # nctkarr_t("mats", "dp", "two, mpw, mpw, ntau, nqibz, nsppol")
            tchi_data = self.tchi_reader.read_variable("mats")[0, tchi_iq, :, chi_ig2, chi_ig1, :]
            tchi_data = tchi_data[:, 0] + 1j * tchi_data[:, 1]

            style = dict(markersize=4)
            ax_re.plot(susc_iwmesh[1:], sus_data[1:].real, ls="--", marker="o", label="AW Re", **style)
            ax_re.plot(tchi_iwmesh, tchi_data[0:].real, ls=":", marker="^", label="minimax Re", **style)
            ax_re.grid(True)

            ax_im.plot(susc_iwmesh[1:], sus_data[1:].imag, ls="--", marker="o", label="AW Im", **style)
            ax_im.plot(tchi_iwmesh, tchi_data[0:].imag, ls=":", marker="^", label="minimax Im", **style)
            ax_im.grid(True)

            tex_g1, tex_g2, tex_qpt = r"${\bf{G}}_1$", r"${\bf{G}}_2$", r"${\bf{q}}$"
            ax_re.set_title(f"{tex_g1}: {g1}, {tex_g2}: {g2}, {tex_qpt}: {qpoint}", fontsize=fontsize)
            ax_im.set_title(f"{tex_g1}: {g1}, {tex_g2}: {g2}, {tex_qpt}: {qpoint}", fontsize=fontsize)

            ax_re.legend(loc="best", fontsize=fontsize, shadow=True)
            ax_im.legend(loc="best", fontsize=fontsize, shadow=True)

            #if i == 0:
            ax_re.set_ylabel(r'$\Re{\chi^0_{\bf{G}_1 \bf{G}_2}(i\omega)}$')
            ax_im.set_ylabel(r'$\Im{\chi^0_{\bf{G}_1 \bf{G}_2}(i\omega)}$')

            if i == len(gpairs) - 1:
                ax_re.set_xlabel(r'$i \omega$ (Ha)')
                ax_im.set_xlabel(r'$i \omega$ (Ha)')

            # The two files may store chi on different meshes.
            # Here we select the min value for comparison purposes.
            w_max = min(susc_iwmesh[-1], tchi_iwmesh[-1]) + 5
            xlims = [0, w_max]
            xlims = [-0.2, 7]

            set_axlims(ax_re, xlims, "x")
            set_axlims(ax_im, xlims, "x")

        if with_title:
            tex1 = r"$\chi^0(i\omega)$"
            tex2 = r"$\chi^0(i\tau) \rightarrow\chi^0(i\omega)$"
            fig.suptitle(f"Comparison between Adler-Wiser {tex1} and minimax {tex2}\nLeft: real part. Right: imag part",
                         fontsize=fontsize)

        return fig

    def expose_qpoints_gpairs(self, qpoint_list, gpairs, exposer="mpl"):
        """
        Plot the Fourier components of the polarizability for a list of q-points and,
        for each q-point, a list of (g, g') pairs.

        Args:
            qpoint_list: List of q-points to consider.
            gpairs: List of (g,g') pairs.
            exposer: "mpl" for matplotlib, "panel" for web interface.
        """
        with Exposer.as_exposer(exposer) as e:
            for qpoint in qpoint_list:
                e(self.plot_qpoint_gpairs(qpoint, gpairs, show=False))
            return e

    @add_fig_kwargs
    def plot_matdiff(self, qpoint, iw_index: int, npwq=101, with_susmat=False,
                     cmap="jet", fontsize=8, spins=(0, 0), **kwargs) -> Figure:
        """
        Plot G-G' matrix with the absolute value of chi_AW and chi_GWR for given q-point and omega index.

        Args:
            qpoint: q-point or q-point index.
            iw_index: Frequency index.
            with_susmat: True to add additional subplot with absolute value of susmat_{g,g'}.
        """
        sus_reader = self.sus_file.reader
        _, sus_iq = sus_reader.find_kpoint_fileindex(qpoint)

        # int reduced_coordinates_plane_waves_dielectric_function(number_of_qpoints_dielectric_function,
        # number_of_coefficients_dielectric_function, number_of_reduced_dimensions) ;
        # SUSC file uses Gamma-centered G-spere so read at iq = 0
        susc_iwmesh = sus_reader.read_value("frequencies_dielectric_function", cmode="c").imag
        susc_gvecs = sus_reader.read_variable("reduced_coordinates_plane_waves_dielectric_function")[0]

        tchi_iq, qibz = self.find_tchim_qpoint(qpoint)
        tchi_iwmesh = self.tchi_reader.read_value("iw_mesh")
        tchi_gvecs = self.tchi_reader.read_variable("gvecs")[tchi_iq]
        #FIXME: Requires new format
        npwq = self.tchi_reader.read_variable("chinpw_qibz")[tchi_iq]
        #npwq = min(npwq, len(tchi_gvecs))
        tchi_gvecs = tchi_gvecs[:npwq]

        # Find indices of tchi_gvecs in susc_gvecs to remap matrix
        ig_tchi2sus = np.array([_find_g(g1, susc_gvecs) for g1 in tchi_gvecs])

        if sus_reader.path.endswith("SUS.nc"):
            sus_var_name = "polarizability"
        elif sus_reader.path.endswith("SCR.nc"):
            sus_var_name = "inverse_dielectric_function"
        else:
            raise ValueError(f"Cannot detect varname for file: {sus_reader.path}")

        # Read full matrix from file, extract smaller submatrix using sus_index, chi_inds
        #
        # number_of_qpoints_dielectric_function, number_of_frequencies_dielectric_function,
        # number_of_spins, number_of_spins, number_of_coefficients_dielectric_function,
        # number_of_coefficients_dielectric_function, complex)
        sus_data = sus_reader.read_variable(sus_var_name)[sus_iq, iw_index, 0, 0]
        sus_data = (sus_data[:,:, 0] + 1j * sus_data[:,:,1]).T.copy()

        # TODO: Very inefficient for large npwq.
        sus_mat = np.zeros((npwq, npwq), dtype=complex)
        for ig2 in range(npwq):
            ig2_sus = ig_tchi2sus[ig2]
            for ig1 in range(npwq):
                ig1_sus = ig_tchi2sus[ig1]
                sus_mat[ig1,ig2] = sus_data[ig1_sus,ig2_sus]
        sus_data = sus_mat

        # nctkarr_t("mats", "dp", "two, mpw, mpw, ntau, nqibz, nsppol")
        tchi_data = self.tchi_reader.read_variable("mats")[0, tchi_iq, iw_index]
        tchi_data = (tchi_data[:,:,0] + 1j * tchi_data[:,:,1]).T.copy()
        tchi_data = tchi_data[:npwq,:npwq]

        # compute and visualize diff mat
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=1, ncols=2 if with_susmat else 1,
                                                squeeze=False)
        ax_list = ax_list.ravel()

        import matplotlib.colors as colors
        mat = data_from_cplx_mode("abs", tchi_data - sus_data)
        ax = ax_list[0]
        im = ax.matshow(mat, cmap=cmap, norm=colors.LogNorm(vmin=mat.min(), vmax=mat.max()))
        fig.colorbar(im, ax=ax)
        qq = qibz[tchi_iq]
        qq_str = "[%.2f, %.2f, %.2f]" % (qq[0], qq[1], qq[2])
        q_info = r"at ${\bf{q}}: %s$" % qq_str
        ww_ev = susc_iwmesh[iw_index] * abu.Ha_eV
        w_info = r", $i\omega$: %.2f eV" % ww_ev
        label = r"$|\chi^0_{AW}({\bf{G}}_1,{\bf{G}}_2) - \chi^0_{GWR}({\bf{G}}_1,{\bf{G}}_2))|$ " + q_info + w_info
        ax.set_title(label, fontsize=10)

        if with_susmat:
            mat = data_from_cplx_mode("abs", sus_data)
            ax = ax_list[1]
            im = ax.matshow(mat, cmap=cmap, norm=colors.LogNorm(vmin=mat.min(), vmax=mat.max()))
            fig.colorbar(im, ax=ax)
            label = r"$|\chi^0_{AW}(\bf{G}_1,\bf{G}_2)|$"
            ax.set_title(label, fontsize=10)

        return fig



#!/usr/bin/env python
"""
This examples shows how to plot the macroscopic dielectric function
computed in the GW code (optdriver 3)
"""
import abipy.data as abidata
from abipy.abilab import abiopen

with abiopen(abidata.ref_file("sio2_SCR.nc")) as ncfile:
    #print(ncfile)
    # The SCR file contains a structure and electron bands in the IBZ.
    # We can thus use the ebands object to plot bands + DOS.
    #edos = ncfile.ebands.get_edos()
    #ncfile.ebands.plot_with_edos(edos, title="KS energies used to compute the SCR file.")
    #ncfile.plot_emacro(title="Macroscopic dielectric function of $SiO_2$ with local-field effects.")
    #ncfile.plot_eelf(title="Electron Energy Loss Function of $SiO_2$")

    #kpoint = [0.0, 0, 0]
    kpoint = [0.5, 0, 0]
    em1 = ncfile.reader.read_wggmat(kpoint)
    print(em1)
    gvec1 = [1, 0, 0]
    gvec2 = [-1, 0, 0]

    em1.plot_freq(gvec1=gvec1, gvec2=gvec2, waxis="real")
    em1.plot_freq(gvec1=[0, 0, 0], gvec2=[1, 0, 0], waxis="imag")

    em1.plot_gg()
    #em1.plot_gg(cplx_mode="abs", wpos=[0], show=False)
    em1.plot_gg(wpos="imag")



"""
FlowModels for ground-state computations.
"""
from __future__ import annotations

import panel as pn

from abc import ABC
#from typing import List
from pydantic import Field, root_validator
from abipy.core.structure import Structure
from abipy.panels.core import ply
#from abipy.panels.viewers import JSONViewer
from abipy.abio.factories import ebands_input
from abipy.abio.inputs import AbinitInput
from abipy.flowtk import TaskManager, Flow
from abipy.flowtk.flows import bandstructure_flow
from abipy.htc.base_models import AbipyModel, MongoConnector, GfsFileDesc
from abipy.htc.structure_models import StructureData
from abipy.htc.flow_models import FlowModel, PresetQuery
from abipy.htc.gs_models import ScfData, NscfData, RelaxData
from abipy.htc.worker import AbipyWorker


class _EbandsFlowModel(FlowModel, ABC):
    """
    This is the base class for models performing band structure calculations.
    It defines the three input files for the SCF/NSCF/DOS calculation
    as well as sub models for storing the output results.

    The input part i.e. the set of abinit variables are customized by the concrete subclass.
    """

    with_gsr: bool = Field(False, description="True if GSR files should be saved in GridFS.")

    #with_den: bool = Field(False, description="True if DEN file should be saved in GridFS.")

    # These inputs can be generated either manually or from meta-parameters + factory functions.

    scf_input: AbinitInput = Field(None, description="Input for the GS-SCF run")

    nscf_input: AbinitInput = Field(None, description="Input for the NSCF run with k-path")

    dos_input: AbinitInput = Field(None, description="Input for the NSCF run with k-mesh (optional")

    #####################################
    # Output (filled in postprocess_flow)
    #####################################

    scf_data: ScfData = Field(None, description="Results produced by the SCF run.")

    #scf_den_gfsd: GfsFileDesc = Field(None, description="Pointer to the DEN file in GridFs. None if file is not stored")

    nscf_kpath_data: NscfData = Field(None, description="Results produced by the NSCF run with a k-path.")

    nscf_kmesh_data: NscfData = Field(None, description="Results produced by the NSCF run with a k-mesh.")

    def postprocess_flow(self, flow: Flow, worker: AbipyWorker) -> None:
        """
        Analyze the flow and fills the model with output results.
        MongoConnector should be used only to insert files in GridFs as the final insertion is done by the caller.
        This function is called by the AbiPy Worker if the flow completed successfully.
        """
        mng_connector = worker.mng_connector
        scf_task = flow[0][0]

        with scf_task.open_gsr() as gsr:
            self.scf_data = ScfData.from_gsr(gsr, mng_connector, self.with_gsr)

        with flow[0][1].open_gsr() as gsr:
            self.nscf_kpath_data = NscfData.from_gsr(gsr, mng_connector, self.with_gsr)

        if self.dos_input is not None:
            with flow[0][2].open_gsr() as gsr:
                self.nscf_kmesh_data = NscfData.from_gsr(gsr, mng_connector, self.with_gsr)

        #if self.with_den:
        #    den_filepath = scf_task.outdir.need_abiext("DEN")
        #    self.scf_den_gfsd = mng_connector.gfs_put_filepath(den_filepath)

    @classmethod
    def get_preset_queries(cls) -> list[PresetQuery]:
        """
        Return list of MongoDB queries typically used to filter documents.
        """
        return [
            PresetQuery.for_large_forces_or_high_pressure("scf_data", cls),
            PresetQuery.for_collinear_magnetic("scf_data", cls),
        ]

    def get_panel_view(self, mng_connector: MongoConnector):
        """
        Return panel object with a view of the model.
        """
        title = self.in_structure_data.get_title()
        structure = self.in_structure_data.structure
        a, b, c = structure.lattice.abc
        alpha, beta, gamma = structure.lattice.angles
        header = f"""\
## {title}

- Lattice lengths: a = {a:.6f}, b = {b:.6f}, c = {c:.6f} Ang
- Lattice angles: α = {alpha:.3f}, β = {beta:.3f}, ɣ = {gamma:.3f} degrees
"""

        if self.is_completed:
            header += f"""\
- Pressure: {self.scf_data.pressure_gpa:.3f} GPa
- Max |Force|: {self.scf_data.max_force_ev_over_ang:.8f} eV/Ang
- Energy: {self.scf_data.energy:.4f} eV
- Energy per atom: {self.scf_data.energy_per_atom:.4f} eV
"""
            ebands_kpath = self.nscf_kpath_data.ebands
            #ebands_kpath = mng_connector.gfs_get_mson_obj(self.nscf_kpath_data.ebands_gfsd)
            plotly_bands = ply(ebands_kpath.plotly(show=False))
        else:
            plotly_bands = pn.pane.Alert(f"Bands are not available because exec_status is `{self.flow_data.exec_status}`")

        return pn.Column(
            #self.in_structure_data.get_panel_view(mongo_connector),
            header,
            pn.Row(
                plotly_bands,
                pn.pane.HTML(self.scf_input._repr_html_()),
            ),
            #"### MongoDB Document",
            #JSONViewer(self.json(), depth=1),
            pn.layout.Divider(),
            sizing_mode="stretch_both",
        )

    #@classmethod
    #def mongo_aggregate_egaps(cls, collection: Collection) -> pd.DataFrame:
        #oids = cls.mng_find_completed_oids(collection)
        #    projection = [
        #        "in_structure_data
        #    }
        #
        #    fund_gap_projection = [
        #        "scf_data.fundamental_gap",
        #        "nscf_kpath_data.fundamental_gap",
        #        "nscf_kmesh_data.fundamental_gap",
        #    ]

        #    direct_gap_projection = {
        #        "scf_data.direct_gap",
        #        "nscf_kpath_data.direct_gap",
        #        "nscf_kmesh_data.direct_gap",
        #    }

        #    projection.extend(fundamental_gap_projection + direct_gap_projection)

        #    oids = cls.mng_find_completed_oids(collection)
        #    rows = []
        #    for oid in oids:
        #        doc = collection.find_one({"_id": oid}, projection)
        #        structure_data = AbipyDecoder().process_decoded(doc["in_structure_data"])
        #        #structure_data = cls.decode(doc["in_structure_data"])
        #        row = structure_data.dict4pandas()
        #        # Here I need a tool to access foo.bar instead of d["foo"]["bar"]
        #        row["fund_gap"] = min(doc[key] for key in fundamental_gap_projection)
        #        row["direct_gap"] = min(doc[key] for key in direct_gap_projection)
        #        rows.append(row)

        #    return pd.DataFrame(rows)


from abipy.htc.protocol import Protocol


class HasProtocol(AbipyModel):
    """
    Mixin class for FlowModels with a protocol object that is used to generate input files
    and pass meta parameters to the flow.
    """

    protocol: Protocol = Field(..., description="Protocol used to generate input files")

    @classmethod
    def from_structure_and_protocol(cls, structure: Structure, protocol: Protocol, **kwargs):
        """
        Build a FlowModel from a structure and the protocol.
        """
        data = dict(
                in_structure_data=StructureData.from_structure(structure),
                pseudos_specs=protocol.pseudos_specs,
                protocol=protocol
        )
        data.update(kwargs)

        return cls(**data)


class EbandsFlowModelWithProtocol(_EbandsFlowModel, HasProtocol):
    """
    This model defines the input arguments used to build a Flow for band structure calculations
    as well as the sub models used to store the final results.
    """

    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
        """
        Build an AbiPy Flow in `workdir` using the input data available in the model and return it.
        """
        structure = self.in_structure_data.structure
        self.scf_input, self.nscf_input = self.protocol.get_ebands_input(structure)

        #pseudos = self.pseudos_specs.get_pseudos()
        #multi = ebands_input(structure, pseudos,
        #                     kppa=self.kppa, nscf_nband=None, ndivsm=self.ndivsm,
        #                     #ecut=6, pawecutdg=None,
        #                     scf_nband=None, accuracy="normal",
        #                     spin_mode=self.spin_mode,
        #                     smearing=self.smearing, charge=self.charge,
        #                     scf_algorithm=None, dos_kppa=self.dos_kppa,
        #                     )

        #multi.set_vars(paral_kgb=self.paral_kgb)

        #if self.dos_kppa is not None:
        #    self.scf_input, self.nscf_input, self.dos_input = multi.split_datasets()
        #else:
        #    self.scf_input, self.nscf_input = multi.split_datasets()

        return bandstructure_flow(workdir, self.scf_input, self.nscf_input)
                                  #dos_inputs=self.dos_input)


class EbandsFlowModelWithParams(_EbandsFlowModel):
    """
    This model defines the input arguments used to build a Flow for band structure calculations
    as well as the sub models used to store the final results.
    """

    ##################
    # Input parameters
    ##################
    kppa: int = Field(1000, description="Defines the sampling used for the SCF run. Defaults to 1000 if not given.")

    spin_mode: str = Field("polarized", description="Spin polarization")

    charge: float = Field(0.0, description="Electronic charge added to the unit cell.")

    smearing: str = Field("fermi_dirac:0.1 eV", description="Smearing technique.")

    # TODO ndivsm < 0 is much better.
    ndivsm: int = Field(2, description="Number of divisions used to sample the smallest segment of the k-path.")

    dos_kppa: int = Field(None,
                          description="Scalar or List of integers with the number of k-points per atom " +
                                      "to be used for the computation of the DOS (None if DOS is not wanted")

    paral_kgb: int = Field(0, description="")

    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
        """
        Build an AbiPy Flow in `workdir` using the input data available in the model and return it.
        """
        pseudos = self.pseudos_specs.get_pseudos()
        structure = self.in_structure_data.structure

        #self.protocol.get_gs_scf_input(structure)
        #self.scf_input, self.nscf_input = self.protocol.get_ebands_input(structure)

        multi = ebands_input(structure, pseudos,
                             kppa=self.kppa, nscf_nband=None, ndivsm=self.ndivsm,
                             #ecut=6, pawecutdg=None,
                             scf_nband=None, accuracy="normal",
                             spin_mode=self.spin_mode,
                             smearing=self.smearing, charge=self.charge,
                             scf_algorithm=None, dos_kppa=self.dos_kppa,
                             )

        multi.set_vars(paral_kgb=self.paral_kgb)

        if self.dos_kppa is not None:
            self.scf_input, self.nscf_input, self.dos_input = multi.split_datasets()
        else:
            self.scf_input, self.nscf_input = multi.split_datasets()

        return bandstructure_flow(workdir, self.scf_input, self.nscf_input,
                                  dos_inputs=self.dos_input)


class EbandsFlowModelWithInputs(_EbandsFlowModel):
    """
    More flexible class for band structure calculations that requires |AbinitInput| objects
    explicitly generated by the user before calling __init__.
    """

    @root_validator
    def check_inputs(cls, values):
        """Enforce the presence of scf_input and nscf_input once the model is created."""
        scf_input, nscf_input = values.get("scf_input"), values.get("nscf_input")
        if scf_input is None or nscf_input is None:
            raise ValueError(f"Building a {cls.__name__} model requires both `scf_input` and `nscf_input`")
        return values

    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
        """
        Build an AbiPy Flow in `workdir` using the input data available in the model and return it.
        """
        return bandstructure_flow(workdir, self.scf_input, self.nscf_input,
                                  dos_inputs=self.dos_input)


#class RelaxFlowModelWithProtocol(FlowModel):
#
#    with_hist: bool = Field(True, description="True if HIST.nc file should be saved in GridFS.")
#
#    with_gsr: bool = Field(False, description="True if GSR.nc file should be saved in GridFS.")
#
#    relax_data: RelaxData = Field(None, description="Results of the structural relaxation.")
#
#    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
#        """
#        Build an AbiPy Flow in `workdir` using the input data available in the model and return it.
#        """
#        pseudos = self.pseudos_specs.get_pseudos()
#        structure = self.in_data_structure.structure
#        relax_input = self.protocol.get_relax_input(structure)
#        return Flow
#
#    def postprocess_flow(self, flow: Flow, worker: AbipyWorker) -> None:
#        """
#        Analyze the flow and fills the model with output results.
#        MongoConnector should be used only to insert files in GridFs as the final insertion is done by the caller.
#        This function is called by the AbiPy Worker if the flow completed successfully.
#        """
#        mng_connector = worker.mng_connector
#        self.relax_data = RelaxData.from_hist_and_gsr_filepaths(
#                                              hist_filepath, gsr_filepath,
#                                              mng_connector, self.with_gsr, self.with_hist)



#!/usr/bin/env python
"""
This script checks that the options in ``manager.yml``, ``scheduler.yml``,
and the environment on the local machine are properly configured.
"""
from __future__ import annotations

import sys
import os
import argparse
import abipy.flowtk as flowtk
import abipy.data as abidata
import abipy.tools.cli_parsers as cli

from monty import termcolor
from monty.termcolor import cprint
from monty.functools import prof_main
from abipy import abilab


def show_managers(options):
    """
    Print table with manager files provided by AbiPy.
    """
    from tabulate import tabulate
    table = []
    root = os.path.join(abidata.dirpath, "managers")
    yaml_paths = [os.path.join(root, f) for f in os.listdir(root) if f.endswith(".yml") and "_manager" in f]
    for path in yaml_paths:
        manager = flowtk.TaskManager.from_file(path)
        hostname = os.path.basename(path).split("_")[0]
        table.append([hostname, manager.qadapter.QTYPE, path])
        if options.verbose > 1:
            print(manager)
    print(tabulate(table, headers=["hostname", "queue-type", "filepath"], tablefmt="rst"))
    return 0


def get_epilog() -> str:
    return """\
Usage example:
    abicheck.py                ==> Test abipy installation and requirements.
    abicheck.py -m             ==> Print table with manager files provided by AbiPy.
    abicheck.py -c             ==> Install Yaml configuration files for manager and scheduler in ~/.abinit/abipy dir.
    abicheck.py --with-flow    ==> Consistency check followed by execution of AbiPy flow.
"""


def get_parser(with_epilog=False):

    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--loglevel', default="ERROR", type=str,
                        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)
    parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                        help='verbose, can be supplied multiple times to increase verbosity.')
    parser.add_argument('--no-colors', default=False, action="store_true", help='Disable ASCII colors.')
    parser.add_argument('--with-flow', default=False, action="store_true", help='Build and run small abipy flow for testing.')
    parser.add_argument("-d", '--flow-dir', type=str, default=None,
                        help='Create AbiPy flow in this directory. If None, a default directory is used,')
    parser.add_argument("-m", '--show-managers', default=False, action="store_true",
                        help="Print table with manager files provided by AbiPy.")

    parser.add_argument("-c", '--create-config', default=False, action="store_true",
                        help="Create yaml configuration files in ~/abinit/.abipy with predefined settings.")
    parser.add_argument("-f", '--force-reinstall', default=False, action="store_true",
                        help="Overwrite yaml configuration files if --create-config and files already exist.")
    return parser


@prof_main
def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg:
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse the command line.
    try:
        options = parser.parse_args()
    except Exception:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    if options.no_colors:
        # Disable colors
        termcolor.enable(False)

    if options.show_managers:
        # Show table with manager configuration files.
        return show_managers(options)

    if options.create_config:
        # Install Yaml configuration files for manager and scheduler.
        abilab.install_config_files(workdir=None, force_reinstall=options.force_reinstall)

    errmsg = abilab.abicheck(verbose=options.verbose)
    if errmsg:
        cprint(errmsg, "red")
        cprint("TIP: Use `--show-managers` to print the manager files provided by AbiPy.\n" +
               "If abicheck.py is failing because it cannot find the manager.yml configuration file",
               "yellow")
        return 2
    else:
        cprint("\nAbipy requirements are properly configured\n", "green")

    if not options.with_flow:
        return 0

    retcode = run_flow(options)
    if retcode == 0:
        cprint("\nTest flow completed successfully\n", "green")

    return retcode


def make_scf_nscf_inputs(paral_kgb=0):
    """Returns two input files: GS run and NSCF on a high symmetry k-mesh."""
    pseudos = abidata.pseudos("14si.pspnc")
    #pseudos = data.pseudos("Si.GGA_PBE-JTH-paw.xml")

    multi = abilab.MultiDataset(structure=abidata.cif_file("si.cif"), pseudos=pseudos, ndtset=2)

    # Global variables
    ecut = 6
    global_vars = dict(ecut=ecut,
                       nband=8,
                       timopt=-1,
                       istwfk="*1",
                       nstep=15,
                       paral_kgb=paral_kgb,
                       iomode=3,
                    )

    if multi.ispaw:
        global_vars.update(pawecutdg=2*ecut)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(ngkpt=[8,8,8], shiftk=[0,0,0])
    multi[0].set_vars(tolvrs=1e-6)

    # Dataset 2 (NSCF run)
    kptbounds = [
        [0.5, 0.0, 0.0], # L point
        [0.0, 0.0, 0.0], # Gamma point
        [0.0, 0.5, 0.5], # X point
    ]

    multi[1].set_kpath(ndivsm=6, kptbounds=kptbounds)
    multi[1].set_vars(tolwfr=1e-12)

    # Generate two input files for the GS and the NSCF run
    scf_input, nscf_input = multi.split_datasets()
    return scf_input, nscf_input


def run_flow(options):
    """Run test flow, return exit code."""
    import tempfile
    workdir = tempfile.mkdtemp(dir=options.flow_dir)
    cprint("Running small flow in workdir: %s" % workdir, "yellow")
    print()

    # Get the SCF and the NSCF input.
    scf_input, nscf_input = make_scf_nscf_inputs()

    # Build the flow.
    flow = flowtk.bandstructure_flow(workdir, scf_input, nscf_input, manager=None)
    flow.build_and_pickle_dump()
    scheduler = flow.make_scheduler()
    retcode = scheduler.start()
    if retcode != 0:
        cprint("Scheduler returned retcode %s" % retcode, "red")
        return retcode

    flow.show_status()
    if not flow.all_ok:
        cprint("Not all tasks in flow reached all_ok", "red")
        retcode = 1

    return retcode


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Band structure with/without SOC
================================

This example shows how to compute the band structure of GaAs with and without spin-orbit term.
We build two BandStructureWork objects inside a loop over nspinor in [1, 2]
nspinor = 1 corresponds to a standard collinear calculation for non-magnetic systems while
nspinor = 2 gives us the non-collinear case with spinor wavefunctions required for the treatment of SOC.
Some of the variables in the input files must be changed depending on the value of nspinor.

We use relativistic NC pseudos made of two terms: scalar pseudo + SOC term.
The SOC term can be deactivated with the input variable `so_psp`.
"""
import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options):
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    structure = abidata.structure_from_ucell("GaAs")
    pseudos = abidata.pseudos("Ga-low_r.psp8", "As_r.psp8")
    num_electrons = structure.num_valence_electrons(pseudos)
    #print("num_electrons:", num_electrons)

    # Usa same shifts in all tasks.
    ngkpt = [4, 4, 4]
    shiftk = [
        [0.5, 0.5, 0.5],
        [0.5, 0.0, 0.0],
        [0.0, 0.5, 0.0],
        [0.0, 0.0, 0.5],
    ]

    # NSCF run on k-path with large number of bands
    kptbounds = [
        [0.5, 0.0, 0.0],  # L point
        [0.0, 0.0, 0.0],  # Gamma point
        [0.0, 0.5, 0.5],  # X point
    ]

    # Initialize the flow.
    flow = flowtk.Flow(options.workdir, manager=options.manager)

    for nspinor in [1, 2]:
        # Multi will contain two datasets (GS + NSCF) for the given nspinor.
        multi = abilab.MultiDataset(structure=structure, pseudos=pseudos, ndtset=2)

        # Global variables.
        multi.set_vars(
            ecut=20,
            nspinor=nspinor,
            nspden=1 if nspinor == 1 else 4,
            so_psp="*0" if nspinor == 1 else "*1",   # Important!
            #paral_kgb=1,
        )

        nband_occ = num_electrons // 2 if nspinor == 1 else num_electrons
        #print(nband_occ)

        # Dataset 1 (GS run)
        multi[0].set_vars(tolvrs=1e-8, nband=nband_occ + 4)
        multi[0].set_kmesh(ngkpt=ngkpt, shiftk=shiftk, kptopt=1 if nspinor == 1 else 4)

        multi[1].set_vars(iscf=-2, nband=nband_occ + 4, tolwfr=1.e-12)
        multi[1].set_kpath(ndivsm=10, kptbounds=kptbounds)

        # Get the SCF and the NSCF input.
        scf_input, nscf_input = multi.split_datasets()

        flow.register_work(flowtk.BandStructureWork(scf_input, nscf_input))

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_gaas_ebands_soc.py -s
#
# then use:
#
#    abirun.py flow_si_ebands ebands --plot -t NscfTask
#
# to analyze (and plot) the electronic bands produced by the NsfTasks of the Flow.
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_gaas_ebands_soc.png?raw=true
#    :alt: Band structure of GaAs without/with SOC.
#
# Alternatively, one can start a GSR robot for the NscfTask with:
#
#        abirun.py flow_gaas_ebands_soc/ robot GSR -t NscfTask
#
# and then plot the two band structure on the same figure with:
#
# .. code-block:: ipython
#
#        In [1]: %matplotlib
#        In [2]: robot.combiplot_ebands()
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_gaas_ebands_soc_combiplot.png?raw=true
#    :alt: Band structure of GaAs without/with SOC on the same graph



#!/usr/bin/env python
r"""
Lobster COHPCAR
===============

This example shows how to analyze the COHPCAR file
produced by Lobster code <http://schmeling.ac.rwth-aachen.de/cohp/>

Use `abiopen.py FILE` with --expose or --print for a command line interface
and --notebook to generate a jupyter notebook.
"""
import os
import abipy.data as abidata

from abipy.abilab import abiopen

dirpath = os.path.join(abidata.dirpath, "refs", "lobster_gaas")
filename = os.path.join(dirpath, "GaAs_COHPCAR.lobster.gz")

# Open the COHPCAR.lobster file (same API for COOPCAR.lobster)
cohp_file = abiopen(filename)
print(cohp_file)

# Plot COHP.
cohp_file.plot(title="GaAs COHP")

# Plot integrated COHP.
cohp_file.plot(what="i", title="GaAs integrated COHP")

# Plot total overlap for all sites listed in `from_site_index`
cohp_file.plot_site_pairs_total(from_site_index=[0], title="COHP total overlap for site index 0")

# Plot partial crystal orbital projections for all sites listed in `from_site_index`
cohp_file.plot_site_pairs_partial(from_site_index=[0], title="COHP with orbital projections from site index 0")



#!/usr/bin/env python
r"""
Non-linear optical properties with DFPT
=======================================

Flow to compute non-linear optical properties with DFPT (static limit).
"""
import sys
import os
import abipy.flowtk as flowtk
import abipy.data as abidata

from abipy import abilab

# NB: This flow requires abinit >= "8.5.2"


def make_scf_input(ecut=10, ngkpt=(8, 8, 8)):
    """
    This function constructs an `AbinitInput` for performing a
    GS-SCF calculation in crystalline AlAs.

    Args:
        ecut: cutoff energy in Ha.
        ngkpt: 3 integers specifying the k-mesh for the electrons.

    Return:
        `AbinitInput` object
    """
    # Initialize the AlAs structure from an internal database. Use the pseudos shipped with AbiPy.
    gs_inp = abilab.AbinitInput(structure=abidata.structure_from_ucell("AlAs"),
                                pseudos=abidata.pseudos("13al.981214.fhi", "33as.pspnc"))

    # Set the value of the Abinit variables needed for GS runs.
    gs_inp.set_vars(
        nband=4,
        ecut=ecut,
        ngkpt=ngkpt,
        nshiftk=1,
        shiftk=[0.0, 0.0, 0.0],
        ixc=7,
        nstep=500,
        iscf=7,
        diemac=5.0,
        toldfe=1.0e-22,
        nbdbuf=0,
        kptopt=1,
    )

    #gs_inp.set_mnemonics(True)
    return gs_inp


def build_flow(options):
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    scf_input = make_scf_input(ecut=10, ngkpt=(6, 6, 6))
    return flowtk.NonLinearCoeffFlow.from_scf_input(options.workdir, scf_input)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



"""
This module defines the FlowModel abstract class,

All HTC Flow-based calculations show inherith from this class and the
extend the implementation by adding new pydantic Fields and methods
that are either required by the abstract interface or by the specification calculation that is being performed.

See FlowModel declaration for further details.
"""
from __future__ import annotations

import traceback
import sys
import logging
#import functools
import pandas as pd

from datetime import datetime
from pprint import pformat
from enum import Enum
from abc import ABC, abstractmethod
from typing import ClassVar, Union, TypeVar, Type, Optional, Any  # , Type,
from pydantic import BaseModel, Field
from bson.objectid import ObjectId
from pymongo.collection import Collection
from abipy.flowtk.tasks import Task
from abipy.flowtk.events import EventReport
from abipy.flowtk import TaskManager, Work, Flow
from abipy.abio.inputs import AbinitInput
from abipy.htc.base_models import (AbipyModel, TopLevelModel, cls2dict, AbipyDecoder,
        QueryResults, MongoConnector, GfsFileDesc)
from abipy.htc.structure_models import StructureData
from abipy.htc.pseudos_models import PseudoSpecs
from abipy.htc.worker import AbipyWorker

logger = logging.getLogger(__name__)


class PresetQuery(BaseModel):
    """
    A MongoDB preset query that can be used in GUIs or command line interfaces
    to provide ready-to-use values for queries and projections.

    This model is generated by the FlowModel in the `get_preset_queries` method using class methods
    Indeed only the FlowModel is aware of the different submodels, their name, type
    and the physical results stored in the Field.
    """

    query: dict = Field(..., description="MongoDB filter")

    projection: Union[None, list, dict] = Field(..., description="MongoDB projection")

    info: str = Field(..., description="Human-reable description of the query")

    def to_string(self, title: Optional[str] = None, verbose: int = 0) ->  str:
        """
        String representation.
        """
        lines = []
        app = lines.append
        if title is None:
            app(f"=== {self.__class__.__name__} ===")
        else:
            app(title)
        app(f"{self.info}")
        app("MongoDB filter:")
        app(pformat(self.query))
        #if verbose:
        app("MongoDB projection:")
        app(pformat(self.projection))
        app("")

        return "\n".join(lines)

    def __str__(self) -> str:
        return self.to_string()

    def find(self, collection: Collection, **kwargs):
        return collection.find(self.query, self.projection, **kwargs)

    def get_dataframe(self, collection: Collection, try_short_keys: bool = False, **kwargs) -> pd.DataFrame:
        """
        Args:
            shor_keys:
                WARNING don't use this option if you need reproducible results
        """

        def can_use_short_colnames(names):
            """
            We usually receive column names of the form `foo.bar`, `foo.hello`.
            This function detects if it possible to use the string after the last "." as column name.
            without incurring in duplication.
            """
            d = {}
            for name in names:
                i = name.rfind(".")
                if i != -1 and i != len(name) - 1:
                    d[name] = name[i + 1:]
                else:
                    d[name] = name

            can_shorten = len(set(d.values())) == len(names)
            return can_shorten, d

        #TODO Add limit
        # https://towardsdatascience.com/all-pandas-json-normalize-you-should-know-for-flattening-json-13eae1dfb7dd
        cursor = self.find(collection, **kwargs)
        df = pd.json_normalize(list(cursor))

        if try_short_keys:
            can_use, mapper = can_use_short_colnames(df.keys())
            if can_use:
                df = df.rename(columns=mapper)

        return df

    ##############
    # Constructors
    ##############

    @classmethod
    def for_large_forces_or_high_pressure(cls, root_name: str, model_cls: Type[TopLevelModel],
                                          force_ev_ang: float = 1e-4, p_gpa: float = 2.0) -> PresetQuery:
        """
        Find documents with large forces or large pressure.

        Args:
            root_name: Usually `scf_data`. Uses results stored in a ScfData submodel
        """
        #assert "." not in root_name "Nested fields are not supported"
        model_field = model_cls.__fields__[root_name]

        # Build query strings from root_name.
        abs_pressure_gpa = f"{root_name}.abs_pressure_gpa"
        max_force = f"{root_name}.max_force_ev_over_ang"

        #assert model_field.type_ is ScfData
        #for f in [abs_pressure, max_force]:
        #    assert f in model_field.type_

        return cls(
                query={"$or": [
                    {abs_pressure_gpa: {"$gt": p_gpa}},
                    {max_force: {"$gt": force_ev_ang}}]
                },
                projection=["in_structure_data.formula_pretty", abs_pressure_gpa, max_force],
                info=f"Filter `{model_cls.__name__}` documents with pressure > {p_gpa} GPa "
                     f"or max |force| > {force_ev_ang} eV/Ang",
                )

    @classmethod
    def for_collinear_magnetic(cls, root_name: str, model_cls: Type[TopLevelModel]) -> PresetQuery:
        """
        Find documents with non-zero collinear magnetization"

        Args:
            root_name: Usually `scf_data`. Uses results stored in a ScfData submodel
        """
        #assert "." not in root_name "Nested fields are not supported"
        #model_field = model_cls.__fields__[root_name]

        # Build query strings from root_name.
        key = f"{root_name}.collinear_mag"

        return cls(
                query={key: {"$nin": [None, 0.0]}},
                projection=["in_structure_data.formula_pretty", key],
                info=f"Filter `{model_cls.__name__}` documents with non-zero collinear magnetization"
                )

    @classmethod
    def for_semiconductors(cls, root_name: str, model_cls: Type[TopLevelModel],
                           min_gap_ev=0.1) -> PresetQuery:
        """
        Find documents with non-zero collinear magnetization"

        Args:
            root_name: Usually `nscf_data`. Uses results stored in a NscfData submodel
        """
        #assert "." not in root_name "Nested fields are not supported"
        #model_field = model_cls.__fields__[root_name]

        # Build query strings from root_name.
        key = f"{root_name}.fundamental_gap_ev"

        return cls(
                query={key: {"$gt": min_gap}},
                projection=["in_structure_data.formula_pretty", key,
                            f"{root_name}.direct_gap_ev",
                ],
                info=f"Filter `{model_cls.__name__}` documents with non-zero collinear magnetization"
                )


class _NodeData(AbipyModel):

    node_id: str = Field(..., description="Node identifier")

    node_class: str = Field(..., description="Class of the Node")

    status: str = Field(..., description="Status of the Node")

    @classmethod
    def get_data_for_node(cls, node) -> dict:
        return dict(
            node_id=node.node_id,
            node_class=node.__class__.__name__,
            status=str(node.status),
        )


class TaskData(_NodeData):
    """
    Data Model associated to an AbiPy |Task|.
    """

    input: AbinitInput = Field(..., description="Abinit input object")

    #name: str: Field(..., description="Name of the Task e.g. w0_t0")

    num_warnings: int = Field(..., description="Number of warnings found the in log file")

    num_errors: int = Field(..., description="Number of errors")

    num_comments: int = Field(..., description="Number of comments found in the log file")

    num_restarts: int = Field(..., description="Number of restarts performed by AbiPy")

    num_corrections: int = Field(..., description="Number of correction performed by AbiPy")

    report: EventReport = Field(..., description="Number of warnings")

    #abinit_abort_file_str: str = Field(None, description="Number of warnings")

    mpi_procs: int = Field(..., description="Number of MPI processes used by the Task")

    omp_threads: int = Field(..., description="Number of OpenMP threads. -1 if not used.")

    #cpu_time: float = Field(..., description="CPU-time in seconds")
    #wall_time: float = Field(..., description="Wall-time in seconds")
    #history: list[str] = Field(..., description="Task history. List of strings logged by AbiPy")

    out_gfsd: GfsFileDesc = Field(None, description="Metadata needed to retrieve the output file from GridFS.")

    log_gfsd: GfsFileDesc = Field(None, description="Metadata needed to retrieve the log file from GridFS.")

    @classmethod
    def from_task(cls, task: Task, mng_connector: MongoConnector, with_out: bool, with_log: bool) -> TaskData:
        """
        Build the model from an AbiPy |Task|
        """
        data = cls.get_data_for_node(task)
        data["input"] = task.input

        # TODO: Handle None!
        report = task.get_event_report()
        data["report"] = report
        for a in ("num_errors", "num_comments", "num_warnings"):
            data[a] = getattr(report, a)

        data.update(dict(
            #"%.1f" % task.mem_per_proc.to("Gb"))))
            ##(task.num_launches,
            num_restarts=task.num_restarts,
            num_corrections=task.num_corrections,
            #stime,
            #task.node_id]))
            #task.status
            mpi_procs=task.mpi_procs,
            omp_threads=task.omp_threads,
        ))

        if with_out:
            data["out_gfsd"] = mng_connector.gfs_put_filepath(task.output_file.path)

        if with_log:
            data["log_gfsd"] = mng_connector.gfs_put_filepath(task.log_file.path)

        return cls(**data)


class WorkData(_NodeData):
    """
    Data Model associated to an AbiPy |Work|
    """

    tasks_data: list[TaskData] = Field(..., description="List of TaskData")

    @classmethod
    def from_work(cls, work: Work, mng_connector: MongoConnector, with_out: bool, with_log: bool) -> WorkData:
        """
        Build the model from a AbiPy |Work| instance.
        """
        data = cls.get_data_for_node(work)
        data["tasks_data"] = [TaskData.from_task(task, mng_connector, with_out, with_log) for task in work]

        return cls(**data)

    def summarize(self, verbose: int = 0, stream=sys.stdout) -> None:
        def p(string: str) -> None:
            print(string, file=stream)

        df = self.get_dataframe()
        p(df)

    def get_dataframe(self) -> pd.DataFrame:
        rows = []
        anames = [
            "num_warnings", "num_errors", "num_comments", "num_restarts", "num_corrections",
            "mpi_procs", "omp_threads"
            #"cpu_time", "wall_time"

        ]

        for task_data in self.tasks_data:
            rows.append({k: getattr(task_data, k) for k in anames})

        return pd.DataFrame(rows)


class ExecStatus(str, Enum):
    """
    Possible status of the execution flow.
    """
    init = "Initialized"
    built = "Built"
    errored = "Errored"
    completed = "Completed"

    #@classmethod
    #def is_member(cls, key):
    #    return key in cls.__members__


class FlowData(AbipyModel):
    """
    Submodel storing the status the Flow and additional metadata related to the execution.
    Users do not need to interact with FlowData explicitly as this operation is delegated to the AbipyWorker.
    """
    exec_status: ExecStatus = ExecStatus.init

    worker_name: str = Field(None, description="The name of the AbiPy worker executing this Flow.")

    worker_hostname: str = Field(None, description="The name of the machine executing this Flow.")

    workdir: str = Field(None, description="The directory where the Flow is executed.")

    created_at: datetime = Field(
        default_factory=datetime.utcnow,
        description="Timestamp for when this document was first created (UTC)",
    )

    completed_at: datetime = Field(
        None,
        description="Timestamp for the most recent calculation update for this property (UTC)",
    )

    works_data: list[WorkData] = Field([], description="List of WorkData objects")

    tracebacks: list[str] = Field([], description="List of exception tracebacks produced by the AbiPy Worker")

    # Private class attributes
    #_aggregation_class_method_names: list[str] = []

    def summarize(self, verbose: int = 0, stream=sys.stdout) -> None:
        """

        Args:
            verbose: Verbosity level.
            stream: File-like object, Default: sys.stdout
        """
        def p(string: str) -> None:
            print(string, file=stream)

        p(f"FlowModel: {self.__class__.__name__}")
        p(f"AbiPyWorker: {self.worker_name}@{self.worker_hostname}")
        p(f"Workdir: {self.workdir}")
        p(f"Created at:   {self.created_at} (UTC)")
        p(f"Completed at: {self.completed_at} (UTC)")
        p(f"ExecStatus: {self.exec_status.value}")

        if self.tracebacks:
            p("WARNING racebacks is not empty:")
            for trace in self.tracebacks:
                p(trace)
                p(80 * "*", "\n")
            return

        for work_data in self.works_data:
            work_data.summarize(verbose=verbose, stream=stream)
            #p(work_data)
            #df = work_data.get_dataframe()
            #print(df)


#class __FlowData(_NodeData):
#    """
#    Document associated to an AbiPy |Flow|
#    """
#
#    works_data: list[WorkData] = Field(..., description="")
#
#    @classmethod
#    def from_flow(cls, flow: Flow) -> __FlowData:
#        """
#        Build the model from an AbiPy |Flow|
#        """
#        data = cls.get_data_for_node(flow)
#        data["works_data"] = [WorkData.from_work(work) for work in flow]
#
#        return cls(**data)


FM = TypeVar('FM', bound='FlowModel')


class FlowModel(TopLevelModel, ABC):
    """
    Abstract base class for models associated to a Flow calculation performed by an AbiPy Worker.
    This model implements the businness logic to create a Flow from a MongoDB collection
    post-process the results once the Flow is completed and insert results in the MongoDB collection.

    Users are supposed to use this model to initialize a MongoDB collection with all
    the input arguments that will be used to generate the Flow and provide a concrete
    implementation of:

        - build_flow.
        - postprocess_flow

    The first method receives the input arguments from the MongoDB database
    and use these values to build a flow.

    The second method is invoked by the AbiPy worker when the calculation is completed.
    The function uses the Flow API to fill the ouput part of the model that
    will be then be stored in the database collection.
    The AbipyWorker uses this API to automate calculations.

    NOTES: The flow is assumed to have the same tructure and pseudopotentials.
    """

    flow_data: FlowData = Field(default_factory=FlowData,
                                description="Stores info on the Flow execution status. Set by the AbiPy Worker.")

    in_structure_data: StructureData = Field(..., description="Input structure and associated metadata. Used ")

    pseudos_specs: PseudoSpecs = Field(..., description="Pseudopotential specifications.")

    with_out: bool = Field(False, description="True if main output files should be added to GridFS.")

    with_log: bool = Field(False, description="True if log files should be added to GridFS.")

    #previous_flow_oids: list[ObjectId] = Field(
    #    None,
    #   description="List of ObjectId associated to documents in the some collection that are connected to this one. "
    #               "May be used to implement convergence studies and/or relaxations with different settings "

    # TODO: Add support for:
    #   - Detecting duplicated FlowModels

    #in_hash_value: str = Field(None, description="Hash value computed from the input data.
    #                                              "Used to find duplicated documents in the collection")

    #priority: int = Field(1, description="Priority associated to the calculation. Can be changed by the user at runtime)

    custom_data: dict[str, Any] = Field(None, description="Dictionary with custom entries set by the user.")

    # Private class attributes
    _magic_key: ClassVar[str] = "_flow_model_"

    #_json_for_eqtest: str

    #def __init__(self, **data):
    #    super().__init__(**data)
    #    self.in_hash_value = hash(self.get_json_for_eqtest())

    #def __eq__(self, other) -> bool:
    #   if self.in_hash_value != other.in_hash_value: return False
    #   if seif is other: return True
    #   return self.get_json_for_eqtest() == other.get_json_for_eqtest()
    #   return self._json_for_eqtest == other.json_for_eqtest

    #def get_json_for_eqtest(self) -> str:
    #    # Do not change the order of the keys.
    #    keys = [
    #        "in_structure_data.structure",
    #        "pseudos_specs",
    #    ]
    #    data = {k: getattr(self, k) for k in keys}
    #    return json.dumps(data)


    # TODO: Remove this method. use mongo_connector.add_flow_models

    @classmethod
    def init_collection(cls: Type[FM], collection: Collection) -> ObjectId:
        """
        Initialize a MongoDB collection by inserting a special document containing
        the serialized FlowModel subclass.
        This function must be used when creating the collection of FlowModels for the first time.

        Return: ObiectID of the new document.
        """
        new_dict = cls2dict(cls)

        cnt, oid = 0, None
        for doc in collection.find({cls._magic_key: {"$exists": True}}):
            oid = doc.pop("_id")
            old_dict = doc.get(cls._magic_key)
            cnt += 1
            if new_dict != old_dict:
                raise TypeError(f"Cannot register new Model {new_dict}\n"
                                f"Collection is already associated to the FlowModel {old_dict}")

        if cnt:
            return oid

        return collection.insert_one({cls._magic_key: new_dict}).inserted_id

    @classmethod
    def get_subclass_from_collection(cls, collection: Collection) -> Type[FlowModel]:
        """
        Retrieve the FlowModel subclass from a MongoDB collection.
        The collections should have been initialized previously by calling `init_collection`.

        Return: FlowModel subclass.
        """
        cnt, data = 0, None
        for doc in collection.find({cls._magic_key: {"$exists": True}}):
            _ = doc.pop("_id")
            data = doc.get(cls._magic_key)
            cnt += 1

        if cnt == 0:
            raise RuntimeError(f"Cannot find document with magic key: `{cls._magic_key}` in collection: `{collection.name}`")
        if cnt > 1:
            raise RuntimeError(f"Found {cnt} documents with magic key: `{cls._magic_key}` in collection: `{collection.name}`")

        sub_class = AbipyDecoder().process_decoded(data)

        if sub_class is not cls and cls is not FlowModel:
            raise TypeError(f"Found FlowModel {sub_class} in collection\nwhile it should be {cls}")

        return sub_class

    @classmethod
    def mongo_get_flow_data(cls, oid: ObjectId, collection: Collection, **kwargs) -> FlowData:
        """
        Return the FlowData submodel given the ObjectId of the MongoDB document.
        """
        d = collection.find_one({"_id": ObjectId(oid)}, projection=["flow_data"], **kwargs)
        if d is None:
            raise ValueError(f"Cannot find ObjectID {oid} in {repr(collection)}")
        data = AbipyDecoder().process_decoded(d["flow_data"])
        return FlowData(**data)

    @classmethod
    def mng_find_completed_oids(cls, collection: Collection, **kwargs) -> list[ObjectId]:
        """
        Find all the models in collection that are completed.
        Return: list of ObjectId.
        """
        d = cls.mng_get_status2oids(collection, **kwargs)
        return d[ExecStatus.completed]

    @classmethod
    def mng_get_status2oids(cls, collection: Collection, **kwargs) -> dict[ExecStatus, list[ObjectId]]:
        """
        Return dictionary mapping the Execution status to the list of ObjectId.
        """
        status2oids = {e: [] for e in ExecStatus}
        key = "flow_data.exec_status"
        collection.create_index(key)
        query = {key: {"$exists": True}}
        #if "add_filter" in kwargs: query.update(kwargs.get("add_filter"))

        for doc in collection.find(query, projection=[key]):
            status = ExecStatus(doc["flow_data"]["exec_status"])
            status2oids[status].append(doc["_id"])

        return status2oids

    @classmethod
    def find_runnable_oid_models(cls, collection: Collection, **kwargs) -> list[tuple[ObjectId, FlowModel]]:
        """
        Return list of (oid, model) tuple with the FlowModels in collection that are ready to run.
        """
        # NB: None not only matches itself but also matches “does not exist.”
        # Thus, querying for a key with the value None will return all documents lacking that key
        # hence we have to check that the key is None and $exists:

        # TODO: If something goes wrong when creating the Flow, flow_data is not None and flow_data.status is Init
        #query = {"flow_data": {"$in": [None], "$exists": True}}
        key = "flow_data.exec_status"
        collection.create_index(key)
        query = {key: ExecStatus.init.value}
        items = []
        decoder = AbipyDecoder()
        for doc in collection.find(query, **kwargs):
            oid = doc.pop("_id")
            doc = decoder.process_decoded(doc)
            model = cls(**doc)
            items.append((oid, model))

        return items

    @property
    def is_init(self) -> bool:
        """True if the model is in `init` state"""
        return self.flow_data.exec_status == ExecStatus.init

    @property
    def is_built(self) -> bool:
        """True if the model is in `built` state"""
        return self.flow_data.exec_status == ExecStatus.built

    @property
    def is_completed(self) -> bool:
        """True if the model is in `completed` state"""
        return self.flow_data.exec_status == ExecStatus.completed

    @property
    def is_errored(self) -> bool:
        """True if the model is in `errored` state"""
        return self.flow_data.exec_status == ExecStatus.errored

    @abstractmethod
    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
        """
        Use the data stored in the model to build and return a Flow to the AbipyWorker.
        Must be implemented in the concrete subclass.
        """

    # TODO: Replace arguments with a Context object or introduce Plan object
    # to support multiple collections from which one can fetch input data if needed.

    def build_flow_and_update_collection(self, workdir: str, oid: ObjectId, worker: AbipyWorker) -> Union[Flow, None]:
        """
        This is the API used by the AbiPyWorker to build a Flow from the model.

        It Wraps the ``build_flow`` method implemented by the subclass in order to perform extra operations
        on the model, exception handling and the final insertion in the MongoDB collection.
        """
        collection = worker.mng_connector.get_collection()

        #self.flow_data = flow_data = FlowData()
        flow_data = self.flow_data

        # Set the workdir of the Flow in the model.
        flow_data.workdir = workdir
        flow_data.worker_name = worker.name
        from socket import gethostname
        flow_data.worker_hostname = gethostname()

        try:
            # Call the method provided by the concrete class to build the Flow.
            flow = self.build_flow(workdir, worker)
            flow_data.exec_status = ExecStatus.built
            return flow

        except Exception:
            flow_data.exec_status = ExecStatus.errored
            # Log the exception and add it to tracebacks so to have it in the DB collection.
            exc_string = traceback.format_exc()
            logger.critical(exc_string)
            flow_data.tracebacks.append(exc_string)
            return None

        finally:
            # Update document in the MongoDB collection.
            self.mng_full_update_oid(oid, collection)

    @abstractmethod
    def postprocess_flow(self, flow: Flow, worker: AbipyWorker) -> None:
        """
        Postprocess the Flow and update the model with output results.
        MongoConnector should be used only to insert files in GridFs
        as the final insertion is done by the caller.
        Must be implemented in the concrete subclass.
        """

    def postprocess_flow_and_update_db(self, flow: Flow, oid: ObjectId, worker: AbipyWorker) -> None:
        """
        API used by the AbiPy Worker to postprocess a Flow from the model.
        Wraps the postprocess_flow method implemented in the subclass to add extra
        operations on the model, error handling and finally the insertion in the collection of the model.
        """
        flow_data = self.flow_data
        mng_connector = worker.mng_connector
        collection = mng_connector.get_collection()

        try:
            self.postprocess_flow(flow, worker)

            # Builds works_data entry.
            for work in flow:
                w_data = WorkData.from_work(work, mng_connector, self.with_out, self.with_log)
                flow_data.works_data.append(w_data)

            flow_data.exec_status = ExecStatus.completed

        except Exception:
            flow_data.exec_status = ExecStatus.errored
            flow_data.tracebacks.append(traceback.format_exc())

        finally:
            #print("in postprocessing_flow_and_update_collection with status:", flow_data.exec_status)
            flow_data.completed_at = datetime.utcnow()
            self.mng_full_update_oid(oid, collection)

    @classmethod
    def mng_get_crystal_systems_incoll(cls, collection: Collection, **kwargs) -> list[int]:
        """
        Return list of crystal systems in collection
        kwargs are passed to pymongo `create_index`
        """
        key = "in_structure_data.crystal_system"
        collection.create_index(key, **kwargs)
        return [k for k in collection.distinct(key) if k is not None]

    @classmethod
    def mng_find_by_crystal_system(cls, crystal_system: str, collection: Collection, **kwargs) -> QueryResults:
        """
        Filter documents in the collection according to the crystal system.
        kwargs are passed to mongo.find.
        """
        key = "in_structure_data.crystal_system"
        collection.create_index(key)
        return cls.mng_find({key: crystal_system}, collection, **kwargs)

    @classmethod
    def mng_get_spg_numbers_incoll(cls, collection: Collection, **kwargs) -> list[int]:
        """
        Return list if all space group numbers found in the collection.
        """
        key = "in_structure_data.spg_number"
        collection.create_index(key, **kwargs)
        return [k for k in collection.distinct(key) if k is not None]

    @classmethod
    def mng_find_by_spg_number(cls, spg_number: int, collection: Collection, **kwargs) -> QueryResults:
        """
        Filter documents in the collection according to the space group number.
        kwargs are passed to mongo.find.
        """
        key = "in_structure_data.spg_number"
        collection.create_index(key)
        return cls.mng_find({key: int(spg_number)}, collection, **kwargs)

    @classmethod
    def mng_find_by_formula(cls, formula_pretty: str, collection: Collection, **kwargs) -> QueryResults:
        """
        Filter documents in the collection according to the formula.
        kwargs are passed to pymongo `collection.find`.
        """
        key = "in_structure_data.formula_pretty"
        collection.create_index(key)
        return cls.mng_find({key: formula_pretty}, collection, **kwargs)

    @classmethod
    @abstractmethod
    def get_preset_queries(cls) -> list[PresetQuery]:
        """
        Return list of dictionaries with the MongoDB queries typically used to filter results.
        Empty list if no suggestion is available. Mainly used by the panel-based GUI.
        Must be implemented in the subclass.
        """

    @abstractmethod
    def get_panel_view(self, mng_connector: MongoConnector):
        """Return panel object with a view of the model."""

    @classmethod
    def mng_aggregate_in_struct(cls, collection: Collection) -> pd.DataFrame:
        """
        Aggregate input structures in collection. Return Dataframe.
        """
        rows = []
        for oid in cls.mng_find_completed_oids(collection):
            d = collection.find_one({"_id": oid}, projection=["in_structure_data"])
            structure_data = StructureData(**d["in_structure_data"])
            row = {"_id": oid}
            row.update(structure_data.dict(exclude={"structure"}))
            rows.append(row)

        return pd.DataFrame(rows)

    #@abstractmethod
    #@classmethod
    #def get_aggregate_panel_ui(cls, collection: Collection):
    #    """Return panel object with a view of the model."""

    #def create_indexes(self, collection: Collection):
    #    creates an index if an index of the same specification does not already exist.
    #    from pymongo import IndexModel, ASCENDING, DESCENDING
    #    index1 = IndexModel([("hello", DESCENDING),
    #                         ("world", ASCENDING)], name="hello_world")
    #    index2 = IndexModel([("goodbye", DESCENDING)])
    #    collection.create_indexes([index1, index2])



#!/usr/bin/env python
r"""
Bethe-Salpeter flow with factory functions
==========================================

Calculation of the BSE spectrum with the High-throuhput interface (factory functions).
"""
import sys
import os
import abipy.data as abidata
import abipy.flowtk as flowtk

from abipy import abilab


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Initialize pseudos and Structure.
    pseudos = abidata.pseudos("14si.pspnc")
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))

    kppa = scf_kppa = 1
    nscf_nband = 6
    nscf_ngkpt = [4, 4, 4]
    nscf_shiftk = [0.1, 0.2, 0.3]
    bs_loband = 2
    bs_nband = nscf_nband
    mbpt_sciss = 0.7 * abilab.units.eV_to_Ha
    mdf_epsinf = 12
    ecuteps = 2
    ecut = 12

    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # BSE calculation with model dielectric function.
    multi = abilab.bse_with_mdf_inputs(
        structure, pseudos,
        scf_kppa, nscf_nband, nscf_ngkpt, nscf_shiftk,
        ecuteps, bs_loband, bs_nband, mbpt_sciss, mdf_epsinf,
        ecut=ecut,# pawecutdg=None,
        exc_type="TDA", bs_algo="haydock", accuracy="normal", spin_mode="unpolarized",
        smearing=None)
        #smearing="fermi_dirac:0.1 eV", charge=0.0, scf_algorithm=None)

    work = flowtk.BseMdfWork(scf_input=multi[0], nscf_input=multi[1], bse_inputs=multi[2:])

    flow.register_work(work)
    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())

############################################################################
#
# Run the script with:
#
#     run_ht_si_bsemdf.py -s
#
# then use:
#
#    abirun.py flow_ht_si_bsemdf/ status -v
#
# to get the status of the flow
#
# .. code-block:: bash
#
#        Work #0: <BseMdfWork, node_id=241906, workdir=flow_ht_si_bsemdf/w0>, Finalized=True
#        +--------+-----------+-----------------+--------------+------------+----------+-----------------+----------+-----------+
#        | Task   | Status    | Queue           | MPI|Omp|Gb   | Warn|Com   | Class    | Sub|Rest|Corr   | Time     |   Node_ID |
#        +========+===========+=================+==============+============+==========+=================+==========+===========+
#        | w0_t0  | Completed | 62430@localhost | 1|  1|2.0    | 2|  0      | ScfTask  | (1, 0, 0)       | 0:00:01R |    241907 |
#        +--------+-----------+-----------------+--------------+------------+----------+-----------------+----------+-----------+
#        | w0_t1  | Completed | 62438@localhost | 2|  1|2.0    | 3|  0      | NscfTask | (1, 0, 0)       | 0:00:05R |    241908 |
#        +--------+-----------+-----------------+--------------+------------+----------+-----------------+----------+-----------+
#        | w0_t2  | Completed | 62449@localhost | 2|  1|2.0    | 10|  6     | BseTask  | (1, 0, 0)       | 0:00:17R |    241909 |
#        +--------+-----------+-----------------+--------------+------------+----------+-----------------+----------+-----------+
#
#        all_ok reached
#
# The macroscopic dielectric function produced by the BseTask is stored in the ``out_MDF.nc`` file:
#
# .. code-block:: bash
#
#        abirun.py flow_ht_si_bsemdf/ listext MDF
#
#        Found 1 files with extension `MDF` produced by the flow
#        File                                          Size [Mb]    Node_ID  Node Class
#        ------------------------------------------  -----------  ---------  ------------
#        flow_ht_si_bsemdf/w0/t2/outdata/out_MDF.nc          0.4     241909  BseTask
#
# Open the file with:
#
#       abiopen.py flow_ht_si_bsemdf/w0/t2/outdata/out_MDF.nc
#
# and then inside the ipython terminal, issue:
#
# .. code-block:: ipython
#
#    In [1]: %matplotlib
#    In [2]: abifile.plot_mdfs()
#
# to produce:
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_ht_si_bsemdf.png?raw=true
#    :alt: Band structure of Si in the IBZ and along a k-path
#
